# =============================================================================
# Copyright (c) 2020-2024 Martin F N Cooper
#
# Author: Martin F N Cooper
# License: MIT License
# =============================================================================

"""
AX.25 Port Information

Methods to query for available AX.25 ports, by interface name, port name, or
callsign. Required to obtain, for example, the AX.25 port name for unproto
frames arriving on a given network interface.

.. important:: This implementation relies on the Linux AX.25 stack, and will
    not work on any other platform. An exception is raised if an attempt is
    made to use it on a non-Linux platform.
"""

import fcntl
import platform
import re
import socket
import struct

import ax25

#
# Constants missing from standard Python modules
#

# Socket configuration controls.
SIOCGIFFLAGS  = 0x8913  # get flags
SIOCGIFHWADDR = 0x8927  # Get hardware address

# Maximum length of interface name
IFNAMSIZ = 16

# Standard interface flags (netdevice->flags).
IFF_UP = 0x1

# ARP protocol HARDWARE identifiers.
ARPHRD_AX25 = 3    # AX.25 Level 2

#
# Formats for packing / unpacking structures
#

# ifreq (Get hardware address)
#
# struct ifreq {
#     char ifr_name[IFNAMSIZ]; /* Interface name */
#     struct sockaddr ifr_hwaddr;
# };
_PACK_IFREQ_HWADDR = "16sH14s"

# ifreq (get flags)
#
# struct ifreq {
#    char ifr_name[IFNAMSIZ]; /* Interface name */
#    short           ifr_flags;
# };
_PACK_IFREQ_FLAGS = "16sh"

# struct sockaddr {
#    unsigned short   sa_family;
#    char             sa_data[14];
# };
_PACK_SOCKADDR = "H14s"

#
# RegEx patterns for parsing
#

_AXPORTS_PATTERN = re.compile(r"""
    ^
    (?P<portname>\S+)
    \s+
    (?P<callsign>\S+)
    \s+
    (?P<speed>\d+)
    \s+
    (?P<paclen>\d+)
    \s+
    (?P<window>\d+)
    \s+
    (?P<description>.+)
    $
""", re.VERBOSE)

_AXPORTS_FILE = '/etc/ax25/axports'


class UnsupportedPlatformError(RuntimeError):
    """
    Raised when a Linux-only module is invoked on a non-Linux platform.

    :param str name: Name of the module being used in error.
    """
    def __init__(self, name):
        super().__init__("{} is supported only on Linux".format(name))


class Port:
    """
    Descriptor for an AX.25 port.

    The information in this descriptor is a combination of static data from the
    `axports` file and live interface data from the Linux OS.

    :param str ifname: Network interface name.
    :param str portname: AX.25 port name.
    :param str callsign: Callsign attached to the port.
    :param int speed: Serial line speed between computer and TNC.
    :param int paclen: Maximum packet length to send.
    :param int window: Number of frames to send without an ACK.
    :param str description: User description for this port.
    """
    def __init__(self, ifname, portname, callsign, speed, paclen,
                 window, description):
        self.ifname = ifname
        self.portname = portname
        self.callsign = callsign
        self.speed = int(speed)
        self.paclen = int(paclen)
        self.window = int(window)
        self.description = description

    def __repr__(self):
        v = vars(self)
        kv = [item for k in v for item in (k, v[k])]
        f = ["{}={!r}"] * len(v)
        fs = "{}({})".format(self.__class__.__name__, ", ".join(f))
        return fs.format(*kv)


class PortInfo:
    """
    Centralized access to information about available AX.25 ports.

    The OS is interrogated to determine all currently active AX.25 ports. This
    information is then combined with data from `axports` to form a central
    source of data on available AX.25 ports. Multiple accessors provide
    convenient lookups for different use cases.
    """
    def __init__(self):
        if platform.system() != 'Linux':
            raise UnsupportedPlatformError(__name__)
        self._sock = None
        self._port_info = None

    def load(self):
        """
        Load the data for all available AX.25 ports.

        :returns: `True` if the information was loaded successfully; `False`
            if the information is not available.
        :rtype: bool
        """
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        if not self._sock:
            # Fail if we can't open a socket
            return False

        call_to_if = {}
        ifnames = self._get_interface_names()
        for ifname in ifnames:
            (family, data) = self._get_interface_info(ifname)
            if family != ARPHRD_AX25:
                continue
            if not self._is_interface_up(ifname):
                continue
            af = ax25.Address.unpack(data)
            call_to_if[str(af)] = ifname

        axports = self._get_axport_info()
        for axport in axports:
            axport.ifname = call_to_if[axport.callsign]

        self._port_info = axports
        return True

    def first_port(self):
        """
        Retrieve the first active port configured in the `axports` file.

        The first configured port is frequently used as a default by AX.25
        applications.

        :returns: The first active AX.25 port, or `None` if there are no
            active ports.
        :rtype: Port or None
        """
        return self._port_info[0] if self._port_info else None

    def find_by_ifname(self, ifname):
        """
        Look up port information based on its interface name.

        :param str ifname: Interface name to look up.
        :returns: The port with this interface name, or `None` if there is no
            such port.
        :rtype: Port or None
        """
        return next(
            (p for p in self._port_info if p.ifname == ifname), None)

    def find_by_portname(self, portname):
        """
        Look up port information based on its port name.

        :param str portname: Port name to look up.
        :returns: The port with this port name, or `None` if there is no
            such port.
        :rtype: Port or None
        """
        return next(
            (p for p in self._port_info if p.portname == portname), None)

    def find_by_callsign(self, callsign):
        """
        Look up port information based on its attached callsign.

        :param str callsign: Callsign to look up.
        :returns: The port with this attached callsign, or `None` if there is
            no such port.
        :rtype: Port or None
        """
        return next(
            (p for p in self._port_info if p.callsign == callsign), None)

    def _get_interface_names(self):
        ifnames = []
        with open("/proc/net/dev", "r") as fp:
            lines = fp.read().splitlines()
        # First 2 lines are headers
        for line in lines[2:]:
            ifnames.append(line[:line.find(':')].strip())
        return ifnames

    def _get_interface_info(self, ifname):
        ifreq = fcntl.ioctl(self._sock.fileno(), SIOCGIFHWADDR,
                            struct.pack(_PACK_IFREQ_HWADDR,
                                        bytes(ifname, 'utf-8')[:15], 0, b''))
        (name, family, data) = struct.unpack(_PACK_IFREQ_HWADDR, ifreq)
        return (family, data)

    def _is_interface_up(self, ifname):
        ifreq = fcntl.ioctl(self._sock.fileno(), SIOCGIFFLAGS,
                            struct.pack(_PACK_IFREQ_FLAGS,
                                        bytes(ifname, 'utf-8')[:15], 0))
        (name, flags) = struct.unpack(_PACK_IFREQ_FLAGS, ifreq)
        return flags & IFF_UP

    def _get_axport_info(self):
        axports = []
        with open(_AXPORTS_FILE, 'r') as fp:
            lines = fp.read().splitlines()
        for line in lines:
            if line.startswith('#'):
                continue
            m = _AXPORTS_PATTERN.match(line)
            if m:
                port = Port(None, m['portname'], m['callsign'], m['speed'],
                            m['paclen'], m['window'], m['description'])
                axports.append(port)
        return axports



#!/usr/bin/env python3
# =============================================================================
# Copyright (c) 2022-2024 Martin F N Cooper
#
# Author: Martin F N Cooper
# License: MIT License
# =============================================================================

"""
AX.25 Sockets Example - Connect

A terminal application along the lines of the Linux axcall utility, providing
simple connected mode capability and demonstrating the use of the ax25.socket
package.

Few commands and options are provided, since the focus is on illustrating the
ease with which a simple connected mode application can be constructed on top
of the native Linux AX.25 stack using ax25.socket. In particular, only the
default port is supported in this example application, and vias are not
supported.These are both supported by the ax25.socket package, however, and
would be straightforward to add should someone choose to use this example as
the basis for a more sophisticated application.
"""

import errno
import platform
import queue
import re
import selectors
import socket
import sys
import threading
import tkinter as tk
from tkinter import ttk
import tkinter.scrolledtext

import ax25
import ax25.ports
import ax25.socket


class Connector:
    """
    Encapsulation of all of the socket-related code, simplifying its use from
    the UI. A client of this class need only create an instance with which to
    connect, and then connect and disconnect as appropriate. Incoming data and
    status updates are added to an event queue, which the client should check
    periodically. Outgoing data may be sent at will. Note that instances of
    this class are not reusable.
    """
    _CON_TIMEOUT = 10  # Seconds to wait for connection
    _BUF_LEN = 512  # Buffer length for socket i/o

    def __init__(self):
        self._sock = ax25.socket.Socket()
        self._queue = queue.Queue()
        self._active = True

    @property
    def event_queue(self):
        return self._queue

    def _queue_event(self, kind, data):
        self._queue.put((kind, data))

    def _receive_data(self):
        """
        Retrieves incoming data, using selectors to wait until the socket
        is ready to be read. Data is posted to the queue for the main thread
        to consume.
        """
        with selectors.DefaultSelector() as sel:
            sel.register(self._sock, selectors.EVENT_READ)
            while True:
                events = sel.select()
                for key, mask in events:
                    try:
                        data = self._sock.recv(self._BUF_LEN)
                    except OSError as e:
                        if e.errno != errno.ENOTCONN:  # Connection closed
                            raise
                        self._queue_event('status', 'disconnected')
                        data = None
                    if not data:
                        self._active = False
                        return
                    self._queue_event('data', data)

    def __call__(self):
        """
        Runs the background thread to start the connection and receive all
        incoming data. Status updates and data are posted to a queue, from
        which they can be accessed by the main thread.
        """
        self._sock.bind(self._src)
        self._queue_event('status', 'connecting')
        # The initial connect call will return immediately. To wait for the
        # connection to complete, or timeout, we need to use a selector to
        # wait until the socket is writable.
        self._sock.setblocking(False)
        res = self._sock.connect_ex(self._dst)
        if res == errno.EINPROGRESS:
            with selectors.DefaultSelector() as sel:
                sel.register(self._sock, selectors.EVENT_WRITE)
                events = sel.select(self._CON_TIMEOUT)
                if not events:
                    self._queue_event('status', 'connect-timeout')
                    return
        elif res != 0:
            self._queue_event('status', 'connect-failure')
            return
        self._queue_event('status', 'connected')
        while self._active:
            self._receive_data()

    def connect(self, src, dst, via=None, port=None):
        """
        Called from the main thread to start the connection, spinning up the
        thread that will be used to connect and receive data.
        """
        self._src = src
        self._dst = dst
        self._via = via
        self._port = port
        name = '{}: {} -> {}'.format(port, str(src), str(dst))
        self.thread = threading.Thread(target=self, name=name)
        self.thread.start()

    def disconnect(self):
        """
        Called from the main thread to disconnect, shut down the thread that
        is receiving data, and close the socket.
        """
        self._queue_event('status', 'disconnecting')
        self._active = False
        try:
            self._sock.shutdown(socket.SHUT_RDWR)
            self._sock.close()
        except OSError:
            pass
        finally:
            self._sock = None
        self._queue_event('status', 'disconnected')

    def send(self, data):
        """
        Called from the main thread to send data over the connection.
        """
        if self._sock:
            self._sock.send(data)


class LogPanel(tkinter.scrolledtext.ScrolledText):
    """
    Scrolling text panel that logs the commands typed by the user, the
    responses from the remote system, and any errors encountered. Each
    is displayed in a different color for easy comprehension.
    """
    def __init__(self, master, **kwargs):
        super().__init__(master, bg='black', **kwargs)
        self.tag_config('tag_local', foreground='magenta')
        self.tag_config('tag_remote', foreground='cyan')
        self.tag_config('tag_error', foreground='red')
        self.configure(state=tk.DISABLED)

    def _add_line(self, line, tag):
        self.configure(state=tk.NORMAL)
        self.insert(tk.END, line + '\n', (tag,))
        self.see(tk.END)
        self.configure(state=tk.DISABLED)

    def local(self, line):
        self._add_line(line, 'tag_local')

    def remote(self, line):
        self._add_line(line, 'tag_remote')

    def error(self, line):
        self._add_line(line, 'tag_error')


class InfoBar(ttk.Frame):
    """
    One-line bar that displays the user's callsign, current connection
    status, and a brief indication of how to get help.
    """
    def __init__(self, parent):
        super().__init__(parent)

        self._var_mycall = tk.StringVar()
        lbl_mycall = ttk.Label(
            self, textvariable=self._var_mycall, anchor=tk.W)
        lbl_mycall.grid(column=0, row=0, sticky=tk.W)

        self._var_connection = tk.StringVar()
        lbl_connection = ttk.Label(
            self, textvariable=self._var_connection, anchor=tk.CENTER)
        lbl_connection.grid(column=1, row=0, sticky=tk.E + tk.W)

        self._var_help_info = tk.StringVar()
        lbl_help_info = ttk.Label(
            self, textvariable=self._var_help_info, anchor=tk.E)
        lbl_help_info.grid(column=2, row=0, sticky=tk.E)

        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1, uniform='col')
        self.grid_columnconfigure(1, weight=1, uniform='col')
        self.grid_columnconfigure(2, weight=1, uniform='col')

    def set_mycall(self, call):
        self._var_mycall.set('My Call: {}'.format(call))

    def set_connection(self, call):
        if call:
            connection = 'Connected to {}'.format(call)
        else:
            connection = 'Not connected'
        self._var_connection.set(connection)

    def set_help_info(self, info):
        self._var_help_info.set(info)


class Application(tk.Tk):
    """
    Main application, creating the UI, responding to commands, and managing
    incoming data. Connector instances are created as required.
    """
    _CMD_PATTERN = re.compile(r"""^:(?P<cmd>[a-z])\s*(?:(?P<args>\S.*))?$""")
    _ALARM_PERIOD = 100  # milliseconds to wait between alarms

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._alarm = None
        self._connected = False
        self._connector = None
        self._mycall = None
        self._destination = None
        self._line_remains = ''

        self.title("AX.25 Socket Connected Mode Example")
        self.geometry("800x600")

        # Main application container
        container = ttk.Frame(self)
        container.pack(side=tk.TOP, fill=tk.BOTH, expand=True)

        # Log panel with scrollbar
        self._log = LogPanel(container)
        self._log.pack(side=tk.TOP, fill=tk.BOTH, expand=True, padx=5, pady=5)

        # Info bar
        self._info_bar = InfoBar(container)
        self._info_bar.pack(fill=tk.X, expand=False, padx=5)

        # Text entry field
        self._var_entry = tk.StringVar()
        self._ent_entry = ttk.Entry(container, textvariable=self._var_entry)
        self._ent_entry.pack(
            side=tk.BOTTOM, fill=tk.X, expand=False, padx=5, pady=5)
        self._ent_entry.bind('<Return>', self._handle_line_entry)

        # Use callsign of first port in axports as an initial value for
        # My Call. The user can change this with the :m command.
        port_info = ax25.ports.PortInfo()
        port_info.load()
        self._mycall = port_info.first_port().callsign

        # Set initial values for the info bar
        self._info_bar.set_mycall(self._mycall)
        self._info_bar.set_connection(None)
        self._info_bar.set_help_info('Type \':h\' for help')

        # Capture the close button
        self.protocol("WM_DELETE_WINDOW", self._do_cmd_q)

        # Give the focus to the entry widget
        self._ent_entry.focus_set()

    def _update_status(self, status):
        """
        Handle status updates from the connector, telling the user what's
        going on, and updating connection status.
        """
        reset = False
        if status == 'connecting':
            self._log.local('Connecting to {} ...'.format(self._destination))
        elif status == 'connected':
            self._connected = True
            self._info_bar.set_connection(self._destination)
            self._log.local('Connected to {}'.format(self._destination))
        elif status == 'connect-timeout':
            self._log.error('Connection attempt timed out')
            reset = True
        elif status == 'connect-failure':
            self._log.error('Connection attempt failed')
            reset = True
        elif status == 'disconnecting':
            self._log.local('Disconnecting ...')
        elif status == 'disconnected':
            reset = True
            self._info_bar.set_connection(None)
            self._log.local('Disconnected')
        else:
            self._log.local('Unknown status: {}'.format(status))
        if reset:
            if self._alarm:
                self.after_cancel(self._alarm)
                self._alarm = None
            self._connected = False
            self._destination = None
            self._connector = None

    def _gather_lines(self, data):
        """
        Incoming data generally does not arrive a complete line at a time,
        so we need to gather up data until we have complete lines to send
        to the log.
        """
        if not isinstance(data, str):
            data = data.decode('utf-8')
        parts = data.split('\r')
        if len(self._line_remains):
            parts[0] = self._line_remains + parts[0]
            self._line_remains = ""
        if data[-1] != '\r':
            self._line_remains = parts[-1]
        del parts[-1]
        for part in parts:
            self._log.remote(part)

    def _process_events(self):
        """
        Check for and process any events on the connector queue. These might
        be either status updates or incoming data. This method is called on
        a periodic basis, using an urwid alarm.
        """
        if not self._connector:
            return
        queue = self._connector.event_queue
        while not queue.empty():
            (kind, data) = queue.get()
            if kind == 'status':
                self._update_status(data)
            else:
                self._gather_lines(data)
        self._alarm = self.after(self._ALARM_PERIOD, self._process_events)

    def _do_cmd_c(self, args):
        """ Connect command """
        if self._connected:
            self._log.error('Already connected')
            return
        if len(args) != 1:
            self._log.error('You must provide a callsign to connect to')
            return
        self._destination = args[0]
        self._alarm = self.after(self._ALARM_PERIOD, self._process_events)
        self._connector = Connector()
        self._connector.connect(self._mycall, self._destination)

    def _do_cmd_d(self, args):
        """ Disconnect command """
        if not self._connected:
            self._log.error('Not connected')
            return
        if len(args) > 0:
            self._log.error('No arguments for disconnect')
            return
        self._log.local('Disconnecting')
        self._connector.disconnect()
        if self._alarm:
            self.after_cancel(self._alarm)
            self._alarm = None

    def _do_cmd_h(self, args):
        """ Help command """
        help = [
            'Commands:',
            '  :c <call>   Connect',
            '  :d          Disconnect',
            '  :h          print Help',
            '  :m <call>   set My Call',
            '  :q          Quit'
        ]
        for line in help:
            self._log.local(line)

    def _do_cmd_m(self, args):
        """ My Call command """
        if len(args) != 1:
            self._log.error('Must provide a callsign to set My Call')
            return
        self._mycall = args[0]
        self._info_bar.set_mycall(self._mycall)

    def _do_cmd_q(self, args=None):
        """ Quit command """
        if self._connected:
            self._do_cmd_d([])
        self.destroy()

    def _handle_command(self, command, args):
        """
        A command has a handler method with a prefix of '_do_cmd_'. Look up
        that method and invoke it if it exists; otherwise report an unknown
        command.
        """
        args = args.split() if args else []
        try:
            getattr(self, '_do_cmd_' + command)(args)
        except AttributeError:
            self._log.error('Unknown command: {}'.format(command))

    def _handle_line_entry(self, event):
        """
        Determine whether a line entered by the user is a command, and, if so,
        hand it off to be processed as such. Otherwise, send the entered text
        to the destination, if connected, and echo it in the log.
        """
        text = self._ent_entry.get()
        self._var_entry.set('')
        m = self._CMD_PATTERN.match(text)
        if m:
            self._handle_command(m['cmd'], m['args'])
        else:
            if self._connected:
                self._connector.send((text + '\r').encode('utf-8'))
                self._log.local(text)
            else:
                self._log.error('Not connected')


if platform.system() != 'Linux':
    print("This application is supported only on Linux")
    sys.exit(1)
app = Application()
app.mainloop()



# Configuration file for the Sphinx documentation builder.

# -- Path setup --------------------------------------------------------------

import os
import sys
sys.path.insert(0, os.path.abspath('..'))

import ax25


# -- Project information -----------------------------------------------------

project = 'PyHam AX.25'
copyright = '2024, Martin F N Cooper. All rights reserved'
author = 'Martin F N Cooper'
release = ax25.__version__
version = release


# -- General configuration ---------------------------------------------------

extensions = [
    'autoapi.extension'
]
autoapi_dirs = ['../ax25']
autoapi_options = [
    'members',
    'show-inheritance',
    'show-module-summary',
    'imported-members'
]

templates_path = ['_templates']

rst_prolog = """
.. meta::
   :author: Martin F N Cooper
   :description: A set of Python modules for working with AX.25 packets in an
      amateur packet radio environment.
"""


# -- Options for HTML output -------------------------------------------------

html_theme = 'sphinx_rtd_theme'
html_theme_options = {
    'prev_next_buttons_location': 'none'
}
html_show_sourcelink = False



# =============================================================================
# Copyright (c) 2020-2024 Martin F N Cooper
#
# Author: Martin F N Cooper
# License: MIT License
# =============================================================================

from contextlib import nullcontext as does_not_raise
import pytest
import ax25


@pytest.mark.parametrize(
    "in_dst, in_src, dst_call, dst_ssid, src_call, src_ssid",
    [
        ('WR6ABD-5', 'W1AW', 'WR6ABD', 5, 'W1AW', 0),
        ('WR6ABD-0', 'W1AW-3', 'WR6ABD', 0, 'W1AW', 3),
        (ax25.Address('WR6ABD-5'), ax25.Address('W1AW'),
            'WR6ABD', 5, 'W1AW', 0),
        (ax25.Address('WR6ABD-0'), ax25.Address('W1AW-3'),
            'WR6ABD', 0, 'W1AW', 3),
        (ax25.Address('WR6ABD-5'), 'W1AW', 'WR6ABD', 5, 'W1AW', 0),
        ('WR6ABD-5', ax25.Address('W1AW'), 'WR6ABD', 5, 'W1AW', 0)
    ])
def test_construct_addresses(
        in_dst, in_src, dst_call, dst_ssid, src_call, src_ssid):
    ctl = ax25.Control(ax25.FrameType.UI)
    f = ax25.Frame(in_dst, in_src, control=ctl)
    assert isinstance(f.dst, ax25.Address)
    assert f.dst.call == dst_call
    assert f.dst.ssid == dst_ssid
    assert isinstance(f.src, ax25.Address)
    assert f.src.call == src_call
    assert f.src.ssid == src_ssid


@pytest.mark.parametrize("in_dst, in_src, expectation", [
    ('W1AW', 'WR6ABD', does_not_raise()),
    (42, 'WR6ABD', pytest.raises(TypeError)),
    ('W1AW', False, pytest.raises(TypeError))
])
def test_construct_addresses_error(in_dst, in_src, expectation):
    ctl = ax25.Control(ax25.FrameType.UI)
    with expectation:
        _ = ax25.Frame(in_dst, in_src, control=ctl)


@pytest.mark.parametrize("in_via, expectation", [
    (['W1AW'], does_not_raise()),
    ([ax25.Address('W1AW')], does_not_raise()),
    (['W1AW', 'WR6ABD'], does_not_raise()),
    (('W1AW', 'WR6ABD'), does_not_raise()),
    (42, pytest.raises(TypeError)),
    ([42], pytest.raises(TypeError))
])
def test_construct_via(in_via, expectation):
    ctl = ax25.Control(ax25.FrameType.UI)
    with expectation:
        _ = ax25.Frame('W1AW', 'WR6ABD', control=ctl, via=in_via)


@pytest.mark.parametrize("in_ctl, expectation", [
    (ax25.Control(ax25.FrameType.UI), does_not_raise()),
    (0x13, does_not_raise()),
    ("bad", pytest.raises(TypeError))
])
def test_construct_control(in_ctl, expectation):
    with expectation:
        _ = ax25.Frame('W1AW', 'WR6ABD', control=in_ctl)


@pytest.mark.parametrize("in_ft, in_pid, pid", [
    (ax25.FrameType.I,  0xF0, 0xF0),
    (ax25.FrameType.RR, 0xF0, 0x00),
    (ax25.FrameType.UI, 0xF0, 0xF0)
])
def test_construct_pid(in_ft, in_pid, pid):
    ctl = ax25.Control(in_ft)
    f = ax25.Frame('W1AW', 'WR6ABD', control=ctl, pid=in_pid)
    assert f._pid == pid


@pytest.mark.parametrize("in_ft, in_data, expectation", [
    (ax25.FrameType.I,  None, does_not_raise()),
    (ax25.FrameType.I,  b'abc', does_not_raise()),
    (ax25.FrameType.I,  bytearray(b'abc'), does_not_raise()),
    (ax25.FrameType.I,  bytearray(b'a' * 300), pytest.raises(ValueError)),
    (ax25.FrameType.I,  42, pytest.raises(TypeError)),
    (ax25.FrameType.RR, None, does_not_raise()),
    (ax25.FrameType.RR, b'abc', pytest.raises(ValueError)),
    (ax25.FrameType.UI,  None, does_not_raise()),
    (ax25.FrameType.UI,  b'abc', does_not_raise()),
    (ax25.FrameType.UI,  bytearray(b'a' * 300), pytest.raises(ValueError)),
    (ax25.FrameType.UI,  42, pytest.raises(TypeError))
])
def test_construct_data(in_ft, in_data, expectation):
    ctl = ax25.Control(in_ft)
    with expectation:
        _ = ax25.Frame('W1AW', 'WR6ABD', control=ctl, data=in_data)


def test_getter_dst():
    ctl = ax25.Control(ax25.FrameType.UI)
    f = ax25.Frame('W1AW-3', 'WR6ABD', control=ctl)
    dst = f.dst
    assert isinstance(dst, ax25.Address)
    assert dst.call == 'W1AW'
    assert dst.ssid == 3


def test_getter_src():
    ctl = ax25.Control(ax25.FrameType.UI)
    f = ax25.Frame('W1AW', 'WR6ABD-3', control=ctl)
    src = f.src
    assert isinstance(src, ax25.Address)
    assert src.call == 'WR6ABD'
    assert src.ssid == 3


def test_getter_via():
    in_via = ['W1AW', 'WR6ABD']
    ctl = ax25.Control(ax25.FrameType.UI)
    f = ax25.Frame('W1AW', 'WR6ABD', in_via, ctl)
    via = f.via
    assert isinstance(via, tuple)
    assert len(via) == len(in_via)
    assert str(via[0]) == in_via[0]
    assert str(via[1]) == in_via[1]


def test_getter_control():
    ctl = ax25.Control(ax25.FrameType.UI, poll_final=True)
    f = ax25.Frame('W1AW', 'WR6ABD', control=ctl)
    ctl2 = f.control
    assert ctl2.frame_type == ctl.frame_type
    assert ctl2.poll_final == ctl.poll_final


@pytest.mark.parametrize("in_ft, in_pid, pid", [
    (ax25.FrameType.I,  0xF0, 0xF0),
    (ax25.FrameType.RR, 0xF0, 0x00),
    (ax25.FrameType.UI, 0xF0, 0xF0)
])
def test_getter_pid(in_ft, in_pid, pid):
    ctl = ax25.Control(in_ft)
    f = ax25.Frame('W1AW', 'WR6ABD', control=ctl, pid=in_pid)
    assert f.pid == pid


@pytest.mark.parametrize("in_ft, in_data, data", [
    (ax25.FrameType.I,  None, None),
    (ax25.FrameType.I,  b'abc', b'abc'),
    (ax25.FrameType.I,  bytearray(b'abc'), b'abc'),
    (ax25.FrameType.RR, None, None),
    (ax25.FrameType.UI,  None, None),
    (ax25.FrameType.UI,  b'abc', b'abc'),
    (ax25.FrameType.UI,  bytearray(b'abc'), b'abc')
])
def test_getter_data(in_ft, in_data, data):
    ctl = ax25.Control(in_ft)
    f = ax25.Frame('W1AW', 'WR6ABD', control=ctl, data=in_data)
    assert f.data == data


@pytest.mark.parametrize(
    "in_ft, in_dst, in_src, in_via, in_pid, in_data, expected",
    [
        (ax25.FrameType.RR, 'W1AW', 'WR6ABD', None, 0x00, None,
            b'\xae\x62\x82\xae\x40\x40\x60'
            b'\xae\xa4\x6c\x82\x84\x88\x61'
            b'\x01'),
        (ax25.FrameType.UI, 'W1AW', 'WR6ABD', None, 0xF0,
            b'Hello',
            b'\xae\x62\x82\xae\x40\x40\x60'
            b'\xae\xa4\x6c\x82\x84\x88\x61'
            b'\x03\xf0Hello'),
        (ax25.FrameType.UI, 'W1AW', 'WR6ABD', None, 0xF0,
            bytearray(b'Hello'),
            b'\xae\x62\x82\xae\x40\x40\x60'
            b'\xae\xa4\x6c\x82\x84\x88\x61'
            b'\x03\xf0Hello'),
        (ax25.FrameType.UI, 'W1AW', 'WR6ABD', ['K6EAG', 'KU6S'], 0xF0,
            b'Hello',
            b'\xae\x62\x82\xae\x40\x40\x60'
            b'\xae\xa4\x6c\x82\x84\x88\x60'
            b'\x96\x6c\x8a\x82\x8e\x40\x60'
            b'\x96\xaa\x6c\xa6\x40\x40\x61'
            b'\x03\xf0Hello')
    ])
def test_pack(in_ft, in_dst, in_src, in_via, in_pid, in_data, expected):
    ctl = ax25.Control(in_ft)
    f = ax25.Frame(
        in_dst, in_src, control=ctl, via=in_via, pid=in_pid, data=in_data)
    packed = f.pack()
    assert packed == expected


@pytest.mark.parametrize(
    "in_ft, in_dst, in_src, in_via, in_pid, in_data, expected",
    [
        (ax25.FrameType.RR, 'W1AW', 'WR6ABD', None, 0x00, None,
            b'\xae\x62\x82\xae\x40\x40\x60'
            b'\xae\xa4\x6c\x82\x84\x88\x61'
            b'\x01'),
        (ax25.FrameType.UI, 'W1AW', 'WR6ABD', None, 0xF0,
            b'Hello',
            b'\xae\x62\x82\xae\x40\x40\x60'
            b'\xae\xa4\x6c\x82\x84\x88\x61'
            b'\x03\xf0Hello'),
        (ax25.FrameType.UI, 'W1AW', 'WR6ABD', None, 0xF0,
            bytearray(b'Hello'),
            b'\xae\x62\x82\xae\x40\x40\x60'
            b'\xae\xa4\x6c\x82\x84\x88\x61'
            b'\x03\xf0Hello'),
        (ax25.FrameType.UI, 'W1AW', 'WR6ABD', ['K6EAG', 'KU6S'], 0xF0,
            b'Hello',
            b'\xae\x62\x82\xae\x40\x40\x60'
            b'\xae\xa4\x6c\x82\x84\x88\x60'
            b'\x96\x6c\x8a\x82\x8e\x40\x60'
            b'\x96\xaa\x6c\xa6\x40\x40\x61'
            b'\x03\xf0Hello')
    ])
def test_pack_bytes(in_ft, in_dst, in_src, in_via, in_pid, in_data, expected):
    ctl = ax25.Control(in_ft)
    f = ax25.Frame(
        in_dst, in_src, control=ctl, via=in_via, pid=in_pid, data=in_data)
    packed = bytes(f)
    assert packed == expected


@pytest.mark.parametrize(
    "in_packed, ft, dst, src, via, pid, data",
    [
        (
            b'\xae\x62\x82\xae\x40\x40\x00'
            b'\xae\xa4\x6c\x82\x84\x88\x01'
            b'\x01',
            ax25.FrameType.RR, 'W1AW', 'WR6ABD', None, 0x00, None
        ),
        (
            b'\xae\x62\x82\xae\x40\x40\x00'
            b'\xae\xa4\x6c\x82\x84\x88\x01'
            b'\x03\xf0Hello',
            ax25.FrameType.UI, 'W1AW', 'WR6ABD', None, 0xF0,
            b'Hello'),
        (
            b'\xae\x62\x82\xae\x40\x40\x00'
            b'\xae\xa4\x6c\x82\x84\x88\x00'
            b'\x96\x6c\x8a\x82\x8e\x40\x00'
            b'\x96\xaa\x6c\xa6\x40\x40\x01'
            b'\x03\xf0Hello',
            ax25.FrameType.UI, 'W1AW', 'WR6ABD', ['K6EAG', 'KU6S'], 0xF0,
            b'Hello'
        )
    ])
def test_unpack(in_packed, ft, dst, src, via, pid, data):
    f = ax25.Frame.unpack(in_packed)
    assert f.control.frame_type == ft
    assert str(f.dst) == dst
    assert str(f.src) == src
    if via:
        assert f.via is not None
        assert len(f.via) == len(via)
        for i, v in enumerate(f.via):
            assert str(v) == via[i]
    else:
        assert f.via is None
    assert f.pid == pid
    assert f.data == data



#!/usr/bin/env python3
# =============================================================================
# Copyright (c) 2021-2024 Martin F N Cooper
#
# Author: Martin F N Cooper
# License: MIT License
# =============================================================================

"""
AX.25 Sockets Example - Connect

A terminal application along the lines of the Linux axcall utility, providing
simple connected mode capability and demonstrating the use of the ax25.socket
package.

Few commands and options are provided, since the focus is on illustrating the
ease with which a connected mode application can be constructed on top of the
native Linux AX.25 stack.
"""

import errno
import queue
import re
import selectors
import socket
import threading
import urwid

import ax25
import ax25.ports
import ax25.socket


palette = [
    ('header_text', 'black', 'white'),
    ('log_text_loc', 'dark magenta', 'black'),
    ('log_text_rem', 'dark cyan', 'black'),
    ('log_text_err', 'dark red', 'black'),
    ('entry_text', 'black', 'light gray')
]


class Connector:
    """
    Encapsulation of all of the socket-related code, simplifying its use from
    the UI. A client of this class need only create an instance with which to
    connect, and then connect and disconnect as appropriate. Incoming data and
    status updates are added to an event queue, which the client should check
    periodically. Outgoing data may be sent at will. Note that instances of
    this class are not reusable.
    """
    _CON_TIMEOUT = 10  # Seconds to wait for connection
    _BUF_LEN = 512  # Buffer length for socket i/o

    def __init__(self):
        self._sock = ax25.socket.Socket()
        self._queue = queue.Queue()
        self._active = True

    @property
    def event_queue(self):
        return self._queue

    def _queue_event(self, kind, data):
        self._queue.put((kind, data))

    def _receive_data(self):
        """
        Retrieves incoming data, using selectors to wait until the socket
        is ready to be read. Data is posted to the queue for the main thread
        to consume.
        """
        with selectors.DefaultSelector() as sel:
            sel.register(self._sock, selectors.EVENT_READ)
            while True:
                events = sel.select()
                for key, mask in events:
                    try:
                        data = self._sock.recv(self._BUF_LEN)
                    except OSError as e:
                        if e.errno != errno.ENOTCONN:  # Connection closed
                            raise
                        self._queue_event('status', 'disconnected')
                        data = None
                    if not data:
                        self._active = False
                        return
                    self._queue_event('data', data)

    def __call__(self):
        """
        Runs the background thread to start the connection and receive all
        incoming data. Status updates and data are posted to a queue, from
        which they can be accessed by the main thread.
        """
        self._sock.bind(self._src)  # FIXME: port, etc.
        self._queue_event('status', 'connecting')
        # The initial connect call will return immediately. To wait for the
        # connection to complete, or timeout, we need to use a selector to
        # wait until the socket is writable.
        self._sock.setblocking(False)
        res = self._sock.connect_ex(self._dst)  # FIXME: via, etc.
        if res == errno.EINPROGRESS:
            with selectors.DefaultSelector() as sel:
                sel.register(self._sock, selectors.EVENT_WRITE)
                events = sel.select(self._CON_TIMEOUT)
                if not events:
                    self._queue_event('status', 'connect-timeout')
                    return
        elif res != 0:
            self._queue_event('status', 'connect-failure')
            return
        self._queue_event('status', 'connected')
        while self._active:
            self._receive_data()

    def connect(self, src, dst, via=None, port=None):
        """
        Called from the main thread to start the connection, spinning up the
        thread that will be used to connect and receive data.
        """
        self._src = src
        self._dst = dst
        self._via = via
        self._port = port
        name = '{}: {} -> {}'.format(port, str(src), str(dst))
        self.thread = threading.Thread(target=self, name=name)
        self.thread.start()

    def disconnect(self):
        """
        Called from the main thread to disconnect, shut down the thread that
        is receiving data, and close the socket.
        """
        self._queue_event('status', 'disconnecting')
        self._active = False
        try:
            self._sock.shutdown(socket.SHUT_RDWR)
            self._sock.close()
        except OSError:
            pass
        finally:
            self._sock = None
        self._queue_event('status', 'disconnected')

    def send(self, data):
        """
        Called from the main thread to send data over the connection.
        """
        if self._sock:
            self._sock.send(data)


class Header(urwid.WidgetWrap):
    """
    Header bar for the top of the window. The header shows My Call (left),
    callsign of destination (center), and basic help info (right).
    """
    def __init__(self):
        self._left = urwid.Text('', 'left')
        self._center = urwid.Text('', 'center')
        self._right = urwid.Text('', 'right')
        widget = urwid.AttrMap(urwid.Padding(urwid.Columns([
            urwid.Filler(self._left),
            urwid.Filler(self._center),
            urwid.Filler(self._right)
        ], box_columns=[0, 1, 2]), left=1, right=1), 'header_text')
        super().__init__(widget)

    def set_mycall(self, call):
        self._left.set_text('My Call: {}'.format(call))

    def set_connection(self, call):
        if call:
            connection = 'Connected to {}'.format(call)
        else:
            connection = 'Not connected'
        self._center.set_text(connection)

    def set_info(self, text):
        self._right.set_text(text)


class LogPanel(urwid.WidgetWrap):
    """
    Panel comprising the main part of the terminal window, showing all input,
    output, and error messages (each in different colors). This panel is
    scrollable.
    """
    def __init__(self):
        self._log = urwid.SimpleListWalker([])
        self._list = urwid.ListBox(self._log)
        super().__init__(self._list)

    def _append(self, line, attr):
        text = urwid.Text(line)
        if type(line) is str:
            text = urwid.AttrMap(text, attr)
        self._log.append(text)
        self._list.set_focus(len(self._log) - 1, 'above')

    def local(self, line):
        self._append(line, 'log_text_loc')

    def remote(self, line):
        self._append(line, 'log_text_rem')

    def error(self, line):
        self._append(line, 'log_text_err')


class LineEntry(urwid.Edit):
    """
    An extension of the standard urwid Edit that emits a complete line when
    Enter is pressed, and then clears its contents.
    """
    signals = ['line_entry']

    def keypress(self, size, key):
        super().keypress(size, key)
        if key != 'enter':
            return key
        text = self.get_edit_text().strip()
        self.set_edit_text("")
        self._emit('line_entry', text)


class Application:
    """
    Main application, creating the UI, responding to commands, and managing
    incoming data. Connector instances are created as required.
    """
    _CMD_PATTERN = re.compile(r"""^:(?P<cmd>[cdhmq])\s*(?:(?P<args>\S.*))?$""")
    _ALARM_PERIOD = 0.1  # seconds to wait between alarms

    def __init__(self):
        self._palette = palette
        self._loop = None
        self._alarm = None
        self._connected = False
        self._mycall = None
        self._destination = None
        self._connector = None
        # Use callsign of first port in axports as a default
        port_info = ax25.ports.PortInfo()
        port_info.load()
        self._mycall = port_info.first_port().callsign

    def _create_widgets(self):
        """
        The application comprises three widgets: a header bar at the top, a
        log window filling the bulk of the window, and an entry line at the
        bottom that the user interacts with.
        """
        self._log = LogPanel()
        self._entry = LineEntry(caption="> ", edit_text="")
        urwid.connect_signal(
            self._entry, 'line_entry', self._handle_line_entry)

        self._header = Header()
        self._footer = urwid.AttrMap(self._entry, 'entry_text')
        self._frame = urwid.Frame(
            self._log,
            header=self._header,
            footer=self._footer,
            focus_part='footer')

        self._header.set_mycall(self._mycall)
        self._header.set_connection(None)
        self._header.set_info("Type ':h' for help")

        return self._frame

    def _update_status(self, status):
        """
        Handle status updates from the connector, telling the user what's
        going on, and updating connection status.
        """
        reset = False
        if status == 'connecting':
            self._log.local('Connecting to {} ...'.format(self._destination))
        elif status == 'connected':
            self._connected = True
            self._header.set_connection(self._destination)
            self._log.local('Connected to {}'.format(self._destination))
        elif status == 'connect-timeout':
            self._log.error('Connection attempt timed out')
            reset = True
        elif status == 'connect-failure':
            self._log.error('Connection attempt failed')
            reset = True
        elif status == 'disconnecting':
            self._log.local('Disconnecting ...')
        elif status == 'disconnected':
            reset = True
            self._header.set_connection(None)
            self._log.local('Disconnected')
        else:
            self._log.local('Unknown status: {}'.format(status))
        if reset:
            if self._alarm:
                self._loop.remove_alarm(self._alarm)
                self._alarm = None
            self._connected = False
            self._destination = None
            self._connector = None

    def _process_events(self, loop, data):
        """
        Check for and process any events on the connector queue. These might
        be either status updates or incoming data. This method is called on
        a periodic basis, using an urwid alarm.
        """
        if not self._connector:
            return
        queue = self._connector.event_queue
        while not queue.empty():
            (kind, data) = queue.get()
            if kind == 'status':
                self._update_status(data)
            else:
                lines = data.decode('utf-8').splitlines()
                for line in lines:
                    self._log.remote(line)
        self._alarm = self._loop.set_alarm_in(
            self._ALARM_PERIOD, self._process_events)

    def _do_cmd_c(self, args):
        """ Connect command """
        if self._connected:
            self._log.error('Already connected')
            return
        if len(args) != 1:
            self._log.error('You must provide a callsign to connect to')
            return
        self._destination = args[0]
        self._alarm = self._loop.set_alarm_in(
            self._ALARM_PERIOD, self._process_events)
        self._connector = Connector()
        self._connector.connect(self._mycall, self._destination)

    def _do_cmd_d(self, args):
        """ Disconnect command """
        if not self._connected:
            self._log.error('Not connected')
            return
        if len(args) > 0:
            self._log.error('No arguments for disconnect')
            return
        self._log.local('Disconnecting')
        self._connector.disconnect()
        if self._alarm:
            self._loop.remove_alarm(self._alarm)
            self._alarm = None

    def _do_cmd_h(self, args):
        """ Help command """
        help = [
            'Commands:',
            '  :c <call>   Connect',
            '  :d          Disconnect',
            '  :h          print Help',
            '  :m <call>   set My Call',
            '  :q          Quit'
        ]
        for line in help:
            self._log.local(line)

    def _do_cmd_m(self, args):
        """ My Call command """
        if len(args) != 1:
            self._log.error('Must provide a callsign to set My Call')
            return
        self._mycall = args[0]
        self._header.set_mycall(self._mycall)

    def _do_cmd_q(self, args):
        """ Quit command """
        if self._connected:
            self._do_cmd_d([])
        raise urwid.ExitMainLoop()

    def _handle_command(self, command, args):
        """
        A command has a handler method with a prefix of '_do_cmd_'. Look up
        that method and invoke it if it exists; otherwise report an unknown
        command.
        """
        args = args.split() if args else []
        try:
            getattr(self, '_do_cmd_' + command)(args)
        except AttributeError:
            self._log.error('Unknown command')

    def _handle_line_entry(self, widget, text):
        """
        Determine whether a line entered by the user is a command, and, if so,
        hand it off to be processed as such. Otherwise, send the entered text
        to the destination, if connected, and echo it in the log.
        """
        m = self._CMD_PATTERN.match(text)
        if m:
            self._handle_command(m['cmd'], m['args'])
        else:
            self._log.local(text)
            if self._connected:
                self._connector.send((text + '\r').encode('utf-8'))

    def run(self):
        """ Start the application """
        self._loop = urwid.MainLoop(
            self._create_widgets(),
            palette=self._palette)
        self._loop.run()


app = Application()
app.run()



#!/usr/bin/env python3
# =============================================================================
# Copyright (c) 2020-2024 Martin F N Cooper
#
# Author: Martin F N Cooper
# License: MIT License
# =============================================================================

"""
AX.25 Frame Decoding Example - Listen

A command line application that enables monitoring of AX.25 frames in a manner
similar to that of the 'listen' application available on Linux.

Frames are decoded using the 'ax25' module, and embedded NET/ROM broadcast
data is decoded using the 'ax25.netrom' module. When running under Linux and
on top of the Linux AX.25 protocol stack, the 'ax25.ports' module is used
to identify the incoming port. (Port information is not available via KISS.)

The application supports two different means of acquiring AX.25 packets. On
all platforms, it can connect to software such as Direwolf using KISS, and
receive packets provided to it through that connection. On Linux only, it can
use the native AX.25 protocol stack to receive packets directly. The choice
between these means is made through use of one of the following command line
options:

  -k host:port or --kiss host:port    use KISS via the specified host and port
  -l or --linux                       use Linux AX.25 protocol stack

A subset of the options to the regular Linux 'listen' application is supported,
but as negative options. That is, color, timestamp, and port name are all
enabled by default, but can be turned off if desired. These options are as
follows:

  -c or --no-color    do not color callsigns, etc.
  -p or --no-port     do not include port name
  -t or --no-time     do not include timestamp

Retrieval of AX.25 frames is performed by listen_kiss() or listen_linux(),
depending upon invocation. Decoding of those frames is performed by
print_frame(), and, if the frame includes data, print_frame_data().

Use Ctrl-C to exit the application.
"""

import argparse
import datetime
import platform
import socket
import sys

import ax25
import ax25.netrom

# Missing from 'socket'
ETH_P_AX25 = 2
ETH_P_ALL  = 3


class Color:
    """
    Simple class that provides coloring of values when color is enabled, and
    ignoring of color when it is not. Once an instance is created, the user
    need not be concerned about checking for enabling, or the specific colors
    to be used for different types of values. In addition, values are forced
    to strings here, eliminating the need for the caller to convert first.
    """
    # Color encoding
    CPORT = '\033[92m'
    CCALL = '\033[92m'
    CTIME = '\033[93m'
    COFF  = '\033[0m'

    def __init__(self, enabled):
        if enabled:
            self._color = self._with_color
        else:
            self._color = self._without_color

    def call(self, value):
        return self._color(value, Color.CCALL)

    def port(self, value):
        return self._color(value, Color.CPORT)

    def time(self, value):
        return self._color(value, Color.CTIME)

    def _with_color(self, value, color):
        return color + str(value) + Color.COFF

    def _without_color(self, value, color):
        return str(value)


class Options:
    """
    Simple class that provides 'dot' access to our command line options. The
    boolean values are negated, since the command line options are used to
    disable capabilities, whereas internally we want to know what is enabled.
    """
    def __init__(self, args, host_port):
        self.color = not args.no_color
        self.port = not args.no_port
        self.time = not args.no_time

        self.use_linux = host_port is None
        if host_port:
            self.host = host_port[0]
            self.port = host_port[1]


def get_options():
    """
    Get options from our command line arguments. Only syntax is handled here;
    the legitimacy of options is handled by the caller.
    """
    parser = argparse.ArgumentParser(
        description='Listen for and decode AX.25 packets.',
        epilog='Either --linux or --kiss must be specified.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '-c', '--no-color',
        action='store_true',
        help='do not color callsigns, etc.')
    parser.add_argument(
        '-p', '--no-port',
        action='store_true',
        help='do not include port name')
    parser.add_argument(
        '-t', '--no-time',
        action='store_true',
        help='do not include timestamp')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        '-k', '--kiss',
        metavar='host:port',
        help='use KISS via specified host and port')
    group.add_argument(
        '-l', '--linux',
        action='store_true',
        help='use Linux AX.25 protocol stack')
    args = parser.parse_args()

    # Parse and validate host and port for KISS
    host_port = None
    if args.kiss:
        host_port = args.kiss.split(':')
        if (len(host_port) != 2
                or len(host_port[0]) < 1
                or len(host_port[1]) < 1
                or not host_port[1].isdigit()):
            raise ValueError(
                'Must provide <host_name>:<port_number> for KISS connection')
        host_port[1] = int(host_port[1])

    # Create positive options from negative args
    return Options(args, host_port)


def print_frame_data(frame):
    """
    For I or UI frames, print associated data. Text data (0xF0) is printed
    directly, after allowing for unprintable characters. Any NET/ROM broadcast
    data is decoded and printed in a similar manner to the Linux 'listen'
    application.
    """
    if frame.pid == 0xF0:
        print(frame.data.decode('utf-8', 'replace'))
    elif frame.pid == 0xCF and frame.data[0] == 0xFF:
        try:
            rb = ax25.netrom.RoutingBroadcast.unpack(frame.data)
        except Exception:
            # Ignore malformed NET/ROM data
            return
        print("NET/ROM Routing: {}".format(rb.sender))
        if rb.destinations:
            for d in rb.destinations:
                print("   {!s:>9}   {:<6}   {!s:>9}   {:>3}".format(
                    d.callsign, d.mnemonic, d.best_neighbor, d.best_quality))


def print_frame(frame, port, opts):
    """
    Print the AX.25 frame according to specified options. The output format
    is very similar to that of the Linux 'listen' application. For any I or
    UI frames, print_frame_data() is called to decode any associated data.
    """
    color = Color(opts.color)
    line = ""

    if opts.time:
        line += "{} ".format(
            color.time(datetime.datetime.now().strftime('%H:%M:%S')))
    if opts.port and port:
        line += "{} ".format(color.port(port + ':'))
    line += "fm {} to {}".format(color.call(frame.src), color.call(frame.dst))
    via = frame.via
    if via:
        line += " via "
        line += " ".join([color.call(v) for v in via])

    control = frame.control
    ft = control.frame_type
    line += " ctl {}".format(ft.name)
    if not ft.is_U():
        line += "(nr={})".format(control.recv_seqno)
    if ft.is_I():
        line += "(ns={})".format(control.send_seqno)
    if ft is ax25.FrameType.I or ft is ax25.FrameType.UI:
        line += " pid={:02X} len {}".format(frame.pid, len(frame.data))

    print(line)
    if (ft is ax25.FrameType.I or ft is ax25.FrameType.UI):
        print_frame_data(frame)


def extract_from_kiss(data):
    """
    Very simplistic extraction of data from KISS frame. This is not intended
    to be foolproof, just minimally sufficient to allow the application to
    retrieve and successfully decode the majority of packets.
    """
    if not (data[0] == 0xC0 and data[1] == 0x00 and data[-1] == 0xC0):
        return None
    return data[2:-1].replace(
        b'\xDB\xDD', b'\xDB').replace(b'\xDB\xDC', b'\xC0')


def listen_kiss(opts):
    """
    Listen for AX.25 frames and print their contents. A regular socket is
    opened to the host and port specified by the user.
    """
    try:
        sock = socket.create_connection((opts.host, opts.port))
    except OSError:
        print(f"Unable to connect to {opts.host}:{opts.port}")
        sys.exit(1)
    while True:
        data, _ = sock.recvfrom(1024)
        frame_data = extract_from_kiss(data)
        if frame_data:
            frame = ax25.Frame.unpack(frame_data)
            print_frame(frame, None, opts)


def listen_linux(opts):
    """
    Listen for AX.25 frames and print their contents. A raw socket is opened
    to the underlying Linux AX.25 protocol stack. Note that this is only
    possible when the application is run as root.
    """
    try:
        sock = socket.socket(
            socket.PF_PACKET, socket.SOCK_RAW, socket.htons(ETH_P_AX25))
    except PermissionError:
        print("You must be root to run listen")
        sys.exit(1)

    if opts.port:
        import ax25.ports
        ports = ax25.ports.PortInfo()
        ports.load()
    port = None

    while True:
        data, addr = sock.recvfrom(1024)
        # Discard leading 0x00 byte before actual frame data
        frame = ax25.Frame.unpack(data[1:])
        if opts.port:
            port = ports.find_by_ifname(addr[0]).portname
        print_frame(frame, port, opts)


# Mainline code
if __name__ == "__main__":
    try:
        opts = get_options()
    except ValueError as e:
        print(f"Invalid arguments: {e}")
        sys.exit(1)
    if opts.use_linux and platform.system() != 'Linux':
        print("Linux option is only available on Linux systems")
        sys.exit(1)
    try:
        if opts.use_linux:
            listen_linux(opts)
        else:
            # Enable colored output on Windows if needed
            if opts.color and platform.system() == 'Windows':
                import os
                os.system('color')
            listen_kiss(opts)
    except KeyboardInterrupt:
        print("\b\b\rThanks for listening!")
        sys.exit(0)



# =============================================================================
# Copyright (c) 2020-2024 Martin F N Cooper
#
# Author: Martin F N Cooper
# License: MIT License
# =============================================================================

import pytest
import ax25


@pytest.mark.parametrize("in_type, is_i", [
    (ax25.FrameType.RR,    False),
    (ax25.FrameType.RNR,   False),
    (ax25.FrameType.REJ,   False),
    (ax25.FrameType.SREJ,  False),
    (ax25.FrameType.UI,    False),
    (ax25.FrameType.DM,    False),
    (ax25.FrameType.SABM,  False),
    (ax25.FrameType.DISC,  False),
    (ax25.FrameType.UA,    False),
    (ax25.FrameType.SABME, False),
    (ax25.FrameType.FRMR,  False),
    (ax25.FrameType.XID,   False),
    (ax25.FrameType.TEST,  False),
    (ax25.FrameType.I,     True),
    (ax25.FrameType.S,     False),
    (ax25.FrameType.U,     False)
])
def test_frame_type_i(in_type, is_i):
    assert in_type.is_I() == is_i


@pytest.mark.parametrize("in_type, is_s", [
    (ax25.FrameType.RR,    True),
    (ax25.FrameType.RNR,   True),
    (ax25.FrameType.REJ,   True),
    (ax25.FrameType.SREJ,  True),
    (ax25.FrameType.UI,    False),
    (ax25.FrameType.DM,    False),
    (ax25.FrameType.SABM,  False),
    (ax25.FrameType.DISC,  False),
    (ax25.FrameType.UA,    False),
    (ax25.FrameType.SABME, False),
    (ax25.FrameType.FRMR,  False),
    (ax25.FrameType.XID,   False),
    (ax25.FrameType.TEST,  False),
    (ax25.FrameType.I,     False),
    (ax25.FrameType.S,     True),
    (ax25.FrameType.U,     False)
])
def test_frame_type_s(in_type, is_s):
    assert in_type.is_S() == is_s


@pytest.mark.parametrize("in_type, is_u", [
    (ax25.FrameType.RR,    False),
    (ax25.FrameType.RNR,   False),
    (ax25.FrameType.REJ,   False),
    (ax25.FrameType.SREJ,  False),
    (ax25.FrameType.UI,    True),
    (ax25.FrameType.DM,    True),
    (ax25.FrameType.SABM,  True),
    (ax25.FrameType.DISC,  True),
    (ax25.FrameType.UA,    True),
    (ax25.FrameType.SABME, True),
    (ax25.FrameType.FRMR,  True),
    (ax25.FrameType.XID,   True),
    (ax25.FrameType.TEST,  True),
    (ax25.FrameType.I,     False),
    (ax25.FrameType.S,     False),
    (ax25.FrameType.U,     True)
])
def test_frame_type_u(in_type, is_u):
    assert in_type.is_U() == is_u



# =============================================================================
# Copyright (c) 2020-2024 Martin F N Cooper
#
# Author: Martin F N Cooper
# License: MIT License
# =============================================================================

from contextlib import nullcontext as does_not_raise
import pytest
import ax25


pack_test_data = {
    (ax25.FrameType.I, False, 0, 0): 0x00,
    (ax25.FrameType.I, True, 0, 0): 0x10,
    (ax25.FrameType.I, False, 3, 0): 0x60,
    (ax25.FrameType.I, False, 0, 7): 0x0E,
    (ax25.FrameType.I, False, 3, 7): 0x6E,
    (ax25.FrameType.I, True, 3, 7): 0x7E,
    (ax25.FrameType.RR, False, 0, 0): 0x01,
    (ax25.FrameType.RR, True, 0, 0): 0x11,
    (ax25.FrameType.RR, False, 3, 0): 0x61,
    (ax25.FrameType.RR, False, 0, 7): 0x01,
    (ax25.FrameType.RR, False, 3, 7): 0x61,
    (ax25.FrameType.RR, True, 3, 7): 0x71,
    (ax25.FrameType.UI, False, 0, 0): 0x03,
    (ax25.FrameType.UI, True, 0, 0): 0x13,
    (ax25.FrameType.UI, False, 3, 0): 0x03,
    (ax25.FrameType.UI, False, 0, 7): 0x03,
    (ax25.FrameType.UI, False, 3, 7): 0x03,
    (ax25.FrameType.UI, True, 3, 7): 0x13
}


unpack_test_data = {
    0x00: (ax25.FrameType.I, False, 0, 0),
    0x10: (ax25.FrameType.I, True, 0, 0),
    0x60: (ax25.FrameType.I, False, 3, 0),
    0x0E: (ax25.FrameType.I, False, 0, 7),
    0x6E: (ax25.FrameType.I, False, 3, 7),
    0x7E: (ax25.FrameType.I, True, 3, 7),
    0x01: (ax25.FrameType.RR, False, 0, 0),
    0x11: (ax25.FrameType.RR, True, 0, 0),
    0x61: (ax25.FrameType.RR, False, 3, 0),
    0x71: (ax25.FrameType.RR, True, 3, 0),
    0x03: (ax25.FrameType.UI, False, 0, 0),
    0x13: (ax25.FrameType.UI, True, 0, 0)
}


@pytest.mark.parametrize("in_ft, ft", [
    (ft, ft) for ft in ax25.FrameType])
def test_construct_internal_defaults(in_ft, ft):
    ctl = ax25.Control(in_ft)
    assert ctl._frame_type == ft
    assert not ctl._poll_final
    assert ctl._recv_seqno == 0
    assert ctl._send_seqno == 0


@pytest.mark.parametrize("ft", [ax25.FrameType.RR, ax25.FrameType.I])
@pytest.mark.parametrize("pf", [True, False])
@pytest.mark.parametrize("nr", [0, 1, 2])
@pytest.mark.parametrize("ns", [0, 1, 2])
def test_construct_internal_args(ft, pf, nr, ns):
    ctl = ax25.Control(ft, poll_final=pf, recv_seqno=nr, send_seqno=ns)
    assert ctl._frame_type == ft
    assert ctl._poll_final == pf
    assert ctl._recv_seqno == nr
    assert ctl._send_seqno == ns


@pytest.fixture
def expected_str(request):
    key = (
        request.node.funcargs['ft'],
        request.node.funcargs['pf'],
        request.node.funcargs['nr'],
        request.node.funcargs['ns']
    )
    return '{:02X}'.format(pack_test_data[key])


@pytest.mark.parametrize("ft, pf, nr, ns", pack_test_data.keys())
def test_str(ft, pf, nr, ns, expected_str):
    ctl = ax25.Control(ft, poll_final=pf, recv_seqno=nr, send_seqno=ns)
    assert str(ctl) == expected_str


@pytest.mark.parametrize("ft, pf, nr, ns", pack_test_data.keys())
def test_repr(ft, pf, nr, ns):
    ctl = ax25.Control(ft, poll_final=pf, recv_seqno=nr, send_seqno=ns)
    rep = repr(ctl)
    assert rep.startswith('Control(')
    assert rep.endswith(')')
    assert "type: {}".format(ft.name) in rep
    assert "poll_final: {}".format(pf) in rep
    if ft.is_I() or ft.is_S():
        assert "recv_seqno: {}".format(nr) in rep
    if ft.is_I():
        assert "send_seqno: {}".format(ns) in rep


@pytest.mark.parametrize("in_ft, ft", [
    (ft, ft) for ft in ax25.FrameType])
def test_getter_ft(in_ft, ft):
    ctl = ax25.Control(in_ft)
    assert ctl.frame_type == ft


@pytest.mark.parametrize("ft", [ax25.FrameType.RR, ax25.FrameType.I])
@pytest.mark.parametrize("pf", [True, False])
def test_getter_poll_final(ft, pf):
    ctl = ax25.Control(ft, poll_final=pf)
    assert ctl.poll_final == pf


@pytest.mark.parametrize("in_ft, value, expectation", [
    (ax25.FrameType.I,  0, does_not_raise()),         # I frame ok
    (ax25.FrameType.RR, 0, does_not_raise()),         # S frame ok
    (ax25.FrameType.UI, 0, pytest.raises(TypeError))  # U frame not valid
])
def test_getter_nr(in_ft, value, expectation):
    ctl = ax25.Control(in_ft)
    with expectation:
        assert ctl.recv_seqno == value


@pytest.mark.parametrize("in_ft, value, expectation", [
    (ax25.FrameType.I,  0, does_not_raise()),          # I frame ok
    (ax25.FrameType.RR, 0, pytest.raises(TypeError)),  # S frame not valid
    (ax25.FrameType.UI, 0, pytest.raises(TypeError))   # U frame not valid
])
def test_getter_ns(in_ft, value, expectation):
    ctl = ax25.Control(in_ft)
    with expectation:
        assert ctl.send_seqno == value


@pytest.mark.parametrize("ft", [ax25.FrameType.RR, ax25.FrameType.I])
@pytest.mark.parametrize("pf", [True, False])
def test_setter_poll_final(ft, pf):
    ctl = ax25.Control(ft)
    ctl.poll_final = pf
    assert ctl.poll_final == pf


@pytest.mark.parametrize("in_ft, value, expectation", [
    (ax25.FrameType.I,  1, does_not_raise()),         # I frame ok
    (ax25.FrameType.RR, 1, does_not_raise()),         # S frame ok
    (ax25.FrameType.UI, 1, pytest.raises(TypeError))  # U frame not valid
])
def test_setter_nr(in_ft, value, expectation):
    ctl = ax25.Control(in_ft)
    with expectation:
        ctl.recv_seqno = value


@pytest.mark.parametrize("in_ft, value, expectation", [
    (ax25.FrameType.I,  1, does_not_raise()),          # I frame ok
    (ax25.FrameType.RR, 1, pytest.raises(TypeError)),  # S frame not valid
    (ax25.FrameType.UI, 1, pytest.raises(TypeError))   # U frame not valid
])
def test_setter_ns(in_ft, value, expectation):
    ctl = ax25.Control(in_ft)
    with expectation:
        ctl.send_seqno = value


@pytest.fixture
def expected_pack(request):
    key = (
        request.node.funcargs['ft'],
        request.node.funcargs['pf'],
        request.node.funcargs['nr'],
        request.node.funcargs['ns']
    )
    return pack_test_data[key]


@pytest.mark.parametrize("ft, pf, nr, ns", pack_test_data.keys())
def test_pack(ft, pf, nr, ns, expected_pack):
    ctl = ax25.Control(ft, poll_final=pf, recv_seqno=nr, send_seqno=ns)
    packed = ctl.pack()
    assert packed == expected_pack


@pytest.mark.parametrize("ft, pf, nr, ns", pack_test_data.keys())
def test_pack_int(ft, pf, nr, ns, expected_pack):
    ctl = ax25.Control(ft, poll_final=pf, recv_seqno=nr, send_seqno=ns)
    packed = int(ctl)
    assert packed == expected_pack


def test_pack_err():
    ctl = ax25.Control(ax25.FrameType.UNK)
    with pytest.raises(ValueError):
        _ = ctl.pack()


@pytest.fixture
def expected_unpack(request):
    key = request.node.funcargs['packed']
    return unpack_test_data[key]


@pytest.mark.parametrize("packed", unpack_test_data.keys())
def test_unpack(packed, expected_unpack):
    ctl = ax25.Control.unpack(packed)
    assert ctl.frame_type == expected_unpack[0]
    assert ctl.poll_final == expected_unpack[1]
    # Check nr and ns against internals to avoid access exceptions
    assert ctl._recv_seqno == expected_unpack[2]
    assert ctl._send_seqno == expected_unpack[3]



# =============================================================================
# Copyright (c) 2021-2024 Martin F N Cooper
#
# Author: Martin F N Cooper
# License: MIT License
# =============================================================================

"""
AX.25 Sockets

This module provides socket support for the AX.25 protocol. While Python
defines the AX.25 address family, it does not implement support for it, and
will fail with 'bad family' on any attempt to use it.

A socket class with support for AX.25 addressing is provided, as is a
convenience method for sending unproto messages. Callsigns may be provided
as strings or as ax25.Address instances.

Note: The :meth:`receivefrom_into()` method has not been implemented, because
it would require the creation of a new buffer in any case, negating the benefit
of using an already-existing buffer. Use :meth:`receivefrom()` instead.

.. important:: This implementation relies on the Linux AX.25 stack, and will
    not work on any other platform. An exception is raised if an attempt is
    made to use it on a non-Linux platform.
"""

import ctypes
import ctypes.util
import os
import platform
import socket
import struct

import ax25
import ax25.ports


# Address data structures from C

# typedef struct
#   {
#     char ax25_call[7];
#   }
# ax25_address;
_PACK_AX25_ADDRESS = "7s"
_SIZE_AX25_ADDRESS = struct.calcsize(_PACK_AX25_ADDRESS)

# struct sockaddr_ax25
#   {
#     sa_family_t sax25_family;
#     ax25_address sax25_call;
#     int sax25_ndigis;
#   };
_PACK_SOCKADDR_AX25 = "H" + _PACK_AX25_ADDRESS + "I"
_SIZE_SOCKADDR_AX25 = struct.calcsize(_PACK_SOCKADDR_AX25)

# struct full_sockaddr_ax25
#   {
#     struct sockaddr_ax25 fsa_ax25;
#     ax25_address fsa_digipeater[AX25_MAX_DIGIS];
#   };
_AX25_MAX_DIGIS = 8
_PACK_FULL_SOCKADDR_AX25 = _PACK_SOCKADDR_AX25 + (
    _AX25_MAX_DIGIS * _PACK_AX25_ADDRESS)
_SIZE_FULL_SOCKADDR_AX25 = struct.calcsize(_PACK_FULL_SOCKADDR_AX25)


def _packed_addr(call):
    if not isinstance(call, ax25.Address):
        call = ax25.Address(call)
    return call.pack()


def _make_addr(call, digis):
    ndigis = len(digis) if digis else 0
    if ndigis > _AX25_MAX_DIGIS:
        raise ValueError("Too many digipeaters")
    call_bytes = _packed_addr(call)
    full_addr = bytearray(struct.pack(
        _PACK_SOCKADDR_AX25, socket.AF_AX25, call_bytes, ndigis))
    if ndigis > 0:
        for digi in digis:
            full_addr.extend(_packed_addr(digi))
    return bytes(full_addr.ljust(_SIZE_FULL_SOCKADDR_AX25, b'\x00'))


_libc = None
_ports = None


# Requirements for this module to be used:
#   * Running on Linux
#   * Can load the C library (libc)
#   * Can load the axports file
def _check_requirements():
    global _libc, _ports
    # See if we've already checked
    if _libc and _ports:
        return
    # Check that we're running on Linux
    if platform.system() != 'Linux':
        raise ax25.ports.UnsupportedPlatformError(__name__)
    # Load libc
    libc_name = ctypes.util.find_library('c')
    if not libc_name:
        raise OSError("Cannot load required library")
    try:
        libc = ctypes.CDLL(libc_name, use_errno=True)
    except OSError:
        raise OSError("Cannot load required library")
    # Load port config
    ports = ax25.ports.PortInfo()
    ports.load()
    # All good, so set globals
    _libc = libc
    _ports = ports


class Socket(socket.socket):
    """
    Socket support for the AX.25 address family.

    A subclass of the Python socket class that provides support for the AX.25
    address family. The address family is hardcoded to AF_AX25 here, since
    this class is not designed to be used as anything other than an AX.25
    socket. Those socket methods that require addresses as arguments or return
    addresses are overridden; other socket methods are available as normal.

    :param sock_type: Type of socket to create. The default is
        `socket.SOCK_SEQPACKET`, which is the appropriate type for a connected
        session, the primary use case for this class.
    :type sock_type: socket.SocketKind
    """
    def __init__(self, sock_type=socket.SOCK_SEQPACKET):
        _check_requirements()
        super().__init__(socket.AF_AX25, sock_type, 0)

    def accept(self):
        """
        Wait for an incoming connection.

        Returns a new socket representing the connection, and the callsign of
        the client.

        :returns: A tuple of the socket and the callsign.
        :rtype: (Socket, str)
        """
        addr_buf = ctypes.create_string_buffer(_SIZE_FULL_SOCKADDR_AX25)
        addr_len = ctypes.c_int(len(addr_buf))
        res = _libc.accept(self.fileno(), addr_buf, ctypes.byref(addr_len))
        if res < 0:
            err = ctypes.get_errno()
            raise OSError(err, os.strerror(err))
        # We need to pass in the proto arg while creating the new socket,
        # because there appears to be a bug in copying over the properties,
        # and proto otherwise gets set to a weird and incorrect value.
        new_sock = socket.socket(fileno=res, proto=0)
        fam, call, digis = struct.unpack_from(_PACK_SOCKADDR_AX25, addr_buf)
        return (new_sock, str(ax25.Address.unpack(call)))

    def bind(self, call, port=None):
        """
        Bind the socket to the specified callsign.

        The socket must not already be bound.

        :param call: Callsign to bind to.
        :type call: str or Address
        :param str port: Name of the port to use when binding. Optional.
        :raises ValueError: if an invalid port name is specified.
        :raises OSError: if a bind failure occurs.
        """
        if port:
            port_info = _ports.find_by_portname(port)
            if not port_info:
                raise ValueError("Invalid port")
            port_digis = [port_info.callsign]
        else:
            port_info = _ports.find_by_callsign(str(call))
            if not port_info:
                port_info = _ports.first_port()
                if port_info:
                    port_digis = [port_info.callsign]
                else:
                    port_digis = None
            else:
                port_digis = None
        bind_addr = _make_addr(call, port_digis)
        res = _libc.bind(self.fileno(), bytes(bind_addr), len(bind_addr))
        if res:
            err = ctypes.get_errno()
            raise OSError(err, os.strerror(err))

    def connect(self, call, via=None):
        """
        Connect the socket to a remote station.

        :param call: Callsign of station to connect to.
        :type call: str or Address
        :param via: List of callsigns to use as Via stations. Optional.
        :type via: list[str] or list[Address] or None
        :raises OSError: if a connection failure occurs.
        """
        err = self.connect_ex(call, via)
        if err:
            raise OSError(err, os.strerror(err))

    def connect_ex(self, call, via=None):
        """
        Connect the socket to a remote station.

        Like `connect(call, via)` but return an error code instead of raising
        an exception when a connection failure occurs.

        :param call: Callsign of station to connect to.
        :type call: str or Address
        :param via: List of callsigns to use as Via stations. Optional.
        :type via: list[str] or list[Address] or None
        :returns: `0` on success, or the value of `errno` if a connection
            failure occurs.
        :rtype: int
        """
        bind_addr = _make_addr(call, via)
        res = _libc.connect(self.fileno(), bind_addr, len(bind_addr))
        return ctypes.get_errno() if res else 0

    def recvfrom(self, bufsize):
        """
        Receive data from the socket.

        Like `recv(bufsize)` but also return the sender’s callsign.

        :param int bufsize: Maximum amount of data to be received at once.
        :returns: A tuple of the received data and the sender's callsign.
        :rtype: (bytes, str)
        """
        data_buf = ctypes.create_string_buffer(bufsize)
        addr_buf = ctypes.create_string_buffer(_SIZE_FULL_SOCKADDR_AX25)
        addr_len = ctypes.c_int(len(addr_buf))
        res = _libc.recvfrom(self.fileno(), data_buf, bufsize, 0, addr_buf,
                             ctypes.byref(addr_len))
        if res < 0:
            err = ctypes.get_errno()
            raise OSError(err, os.strerror(err))
        fam, call, digis = struct.unpack_from(_PACK_SOCKADDR_AX25, addr_buf)
        return (bytes(data_buf), str(ax25.Address.unpack(call)))

    def sendto(self, data, call, via=None):
        """
        Send data to the socket.

        The socket should not be connected to a remote socket, since the
        destination is specified by the callsign.

        :param bytes data: Data to be sent.
        :param call: Callsign of destination.
        :type call: str or Address
        :param via: List of callsigns to use as Via stations. Optional.
        :type via: list[str] or list[Address] or None
        :returns: Number of bytes sent.
        :rtype: int
        """
        addr = _make_addr(call, via)
        count = _libc.sendto(
            self.fileno(), data, len(data), 0, addr, len(addr))
        return count


def send_unproto(src, dst, data, via=None, port=None):
    """
    Convenience function for sending an unproto message.

    This function encapsulates the creation and shutdown of the socket around
    the sending of an unproto message.

    :param src: Callsign of sender.
    :type src: str or Address
    :param dst: Callsign of destination.
    :type dst: str or Address
    :param bytes data: Data to send.
    :param via: List of callsigns to use as Via stations. Optional.
    :type via: list[str] or list[Address] or None
    :param str port: Port to use to send this message. Optional.
    :returns: Number of bytes sent.
    :rtype: int
    """
    with Socket(socket.SOCK_DGRAM) as sock:
        sock.bind(src, port)
        count = sock.sendto(data.encode('utf-8'), dst, via)
    return count



# =============================================================================
# Copyright (c) 2020-2024 Martin F N Cooper
#
# Author: Martin F N Cooper
# License: MIT License
# =============================================================================

from contextlib import nullcontext as does_not_raise
import pytest
import ax25.netrom


@pytest.mark.parametrize("call, mnem, neig, qual, expectation", [
    ('W1AW', 'DST1', 'KU6S', 42, does_not_raise()),
    ('W1AW-3', 'DST1', 'KU6S', 42, does_not_raise()),
    ('W1AW', 'DST1', 'KU6S-5', 42, does_not_raise()),
    ('W!AW', 'DST1', 'KU6S', 42, pytest.raises(ValueError)),
    (123.45, 'DST1', 'KU&S', 42, pytest.raises(TypeError)),
    ('W1AW', 'DST1', 'KU&S', 42, pytest.raises(ValueError)),
    ('W1AW', 'DST1234', 'KU6S', 42, pytest.raises(ValueError)),
    ('W1AW', 1234, 'KU6S', 42, pytest.raises(ValueError)),
    ('W1AW', 'DST1', 'KU6S', 442, pytest.raises(ValueError)),
    ('W1AW', 'DST1', 'KU6S', -42, pytest.raises(ValueError))
])
def test_construct(call, mnem, neig, qual, expectation):
    with expectation:
        d = ax25.netrom.Destination(call, mnem, neig, qual)
        assert str(d._callsign) == call
        assert d._mnemonic == mnem
        assert str(d._neighbor) == neig
        assert d._quality == qual


@pytest.mark.parametrize("in_call, in_neig, call, neig", [
    (ax25.Address('W1AW'), ax25.Address('KU6S'), 'W1AW', 'KU6S'),
    (ax25.Address('W1AW-3'), ax25.Address('KU6S'), 'W1AW-3', 'KU6S'),
    (ax25.Address('W1AW', 3), ax25.Address('KU6S'), 'W1AW-3', 'KU6S'),
    (ax25.Address('W1AW'), ax25.Address('KU6S-5'), 'W1AW', 'KU6S-5'),
    (ax25.Address('W1AW'), ax25.Address('KU6S', 5), 'W1AW', 'KU6S-5'),
    (ax25.Address('W1AW'), 'KU6S', 'W1AW', 'KU6S'),
    ('W1AW', ax25.Address('KU6S'), 'W1AW', 'KU6S')
])
def test_construct_addr(in_call, in_neig, call, neig):
    d = ax25.netrom.Destination(in_call, 'DST1', in_neig, 42)
    assert str(d._callsign) == call
    assert d._mnemonic == 'DST1'
    assert str(d._neighbor) == neig
    assert d._quality == 42


@pytest.mark.parametrize("in_call, in_mnem, in_neig, in_qual, s", [
    ('W1AW', 'DST1', 'KU6S', 42, 'W1AW (DST1) -> KU6S (42)'),
    ('W1AW-3', 'DST1', 'KU6S', 42, 'W1AW-3 (DST1) -> KU6S (42)'),
    ('W1AW', 'DST1', 'KU6S-5', 42, 'W1AW (DST1) -> KU6S-5 (42)'),
    ('W1AW', 'MYNODE', 'KU6S', 42, 'W1AW (MYNODE) -> KU6S (42)'),
    ('W1AW', 'MYNODE', 'KU6S', 0, 'W1AW (MYNODE) -> KU6S (0)'),
    ('W1AW', 'MYNODE', 'KU6S', 255, 'W1AW (MYNODE) -> KU6S (255)')
])
def test_str(in_call, in_mnem, in_neig, in_qual, s):
    d = ax25.netrom.Destination(in_call, in_mnem, in_neig, in_qual)
    assert str(d) == s


@pytest.mark.parametrize("in_call, in_mnem, in_neig, in_qual", [
    ('W1AW', 'DST1', 'KU6S', 42),
    ('W1AW-3', 'DST1', 'KU6S', 42),
    ('W1AW', 'DST1', 'KU6S-5', 42),
    ('W1AW', 'MYNODE', 'KU6S', 42),
    ('W1AW', 'MYNODE', 'KU6S', 0),
    ('W1AW', 'MYNODE', 'KU6S', 255)
])
def test_repr(in_call, in_mnem, in_neig, in_qual):
    d = ax25.netrom.Destination(in_call, in_mnem, in_neig, in_qual)
    rep = repr(d)
    assert rep.startswith('Destination(')
    assert rep.endswith(')')
    assert "callsign: {}".format(in_call) in rep
    assert "mnemonic: {}".format(in_mnem) in rep
    assert "neighbor: {}".format(in_neig) in rep
    assert "quality: {}".format(in_qual) in rep


@pytest.mark.parametrize("in_call, in_neig, call, ssid, ncall, nssid", [
    (ax25.Address('W1AW'), ax25.Address('KU6S'), 'W1AW', 0, 'KU6S', 0),
    (ax25.Address('W1AW-3'), ax25.Address('KU6S'), 'W1AW', 3, 'KU6S', 0),
    (ax25.Address('W1AW', 3), ax25.Address('KU6S'), 'W1AW', 3, 'KU6S', 0),
    (ax25.Address('W1AW'), ax25.Address('KU6S-5'), 'W1AW', 0, 'KU6S', 5),
    (ax25.Address('W1AW'), ax25.Address('KU6S', 5), 'W1AW', 0, 'KU6S', 5),
    (ax25.Address('W1AW'), 'KU6S', 'W1AW', 0, 'KU6S', 0),
    ('W1AW', ax25.Address('KU6S'), 'W1AW', 0, 'KU6S', 0)
])
def test_getter_callsign_neighbor(in_call, in_neig, call, ssid, ncall, nssid):
    d = ax25.netrom.Destination(in_call, 'DST1', in_neig, 42)
    assert d.callsign.call == call
    assert d.callsign.ssid == ssid
    assert d.best_neighbor.call == ncall
    assert d.best_neighbor.ssid == nssid


@pytest.mark.parametrize("in_mnem, in_qual, mnem, qual", [
    ('DST1', 0, 'DST1', 0),
    ('DST1', 42, 'DST1', 42),
    ('DST1', 255, 'DST1', 255)
])
def test_getter_mnemonic_quality(in_mnem, in_qual, mnem, qual):
    d = ax25.netrom.Destination('W1AW', in_mnem, 'KU6S', in_qual)
    assert d.mnemonic == mnem
    assert d.best_quality == qual


def test_pack():
    expected = (b'\xae\x62\x82\xae\x40\x40\x60\x44\x53\x54\x31'
                b'\x20\x20\x96\xaa\x6c\xa6\x40\x40\x60\x2a')
    d = ax25.netrom.Destination('W1AW', 'DST1', 'KU6S', 42)
    packed = d.pack()
    assert packed == expected


def test_unpack():
    packed = (b'\xae\x62\x82\xae\x40\x40\x00\x44\x53\x54\x31'
              b'\x20\x20\x96\xaa\x6c\xa6\x40\x40\x00\x2a')
    d = ax25.netrom.Destination.unpack(packed)
    assert str(d.callsign) == 'W1AW'
    assert d.mnemonic == 'DST1'
    assert str(d.best_neighbor) == 'KU6S'
    assert d.best_quality == 42



# =============================================================================
# Copyright (c) 2020-2024 Martin F N Cooper
#
# Author: Martin F N Cooper
# License: MIT License
# =============================================================================

from contextlib import nullcontext as does_not_raise
import pytest
import ax25.netrom


@pytest.mark.parametrize("sender, destinations, expectation", [
    ('MYNODE', None, does_not_raise()),
    ('', None, pytest.raises(ValueError)),
    (42, None, pytest.raises(ValueError)),
    ('MYNODEX', None, pytest.raises(ValueError)),
    ('MYNODE', [], does_not_raise()),
    ('MYNODE', 42, pytest.raises(TypeError)),
    ('MYNODE', [42], pytest.raises(TypeError)),
    ('MYNODE', [ax25.netrom.Destination('W1AW', 'NODE1', 'KU6S', 42)],
        does_not_raise()),
    ('MYNODE',
        [
            ax25.netrom.Destination('W1AW', 'NODE1', 'KU6S', 42),
            ax25.netrom.Destination('WR6ABD', 'NODE1', 'K6EAG', 21)
        ],
        does_not_raise())
])
def test_construct(sender, destinations, expectation):
    with expectation:
        rb = ax25.netrom.RoutingBroadcast(sender, destinations)
        assert rb._sender == sender
        if destinations:
            assert len(rb._destinations) == len(destinations)
            for d in destinations:
                assert d in rb._destinations
        else:
            assert rb._destinations is None


def test_repr_no_dest():
    rb = ax25.netrom.RoutingBroadcast('MYNODE', None)
    rep = repr(rb)
    assert rep.startswith('RoutingBroadcast(')
    assert rep.endswith(')')
    assert "sender: MYNODE" in rep
    assert "destinations:" not in rep


@pytest.mark.parametrize("destinations", [
    ([ax25.netrom.Destination('W1AW', 'NODE1', 'KU6S', 42)]),
    ([
        ax25.netrom.Destination('W1AW', 'NODE1', 'KU6S', 42),
        ax25.netrom.Destination('WR6ABD', 'NODE1', 'K6EAG', 21)
    ])
])
def test_repr(destinations):
    rb = ax25.netrom.RoutingBroadcast('MYNODE', destinations)
    rep = repr(rb)
    assert rep.startswith('RoutingBroadcast(')
    assert rep.endswith(')')
    assert "sender: MYNODE" in rep
    assert "destinations:" in rep
    for d in destinations:
        assert repr(d) in rep


@pytest.mark.parametrize("sender, destinations", [
    ('MYNODE', None),
    ('MYNODE', [ax25.netrom.Destination('W1AW', 'NODE1', 'KU6S', 42)]),
    ('MYNODE',
        [
            ax25.netrom.Destination('W1AW', 'NODE1', 'KU6S', 42),
            ax25.netrom.Destination('WR6ABD', 'NODE1', 'K6EAG', 21)
        ])
])
def test_getters(sender, destinations):
    rb = ax25.netrom.RoutingBroadcast(sender, destinations)
    assert rb.sender == sender
    if destinations:
        assert len(rb.destinations) == len(destinations)
        for d in destinations:
            assert d in rb.destinations
    else:
        assert rb.destinations is None


@pytest.mark.parametrize("sender, destinations, expected", [
    ('MYNODE', None,
        b'\xffMYNODE'),
    ('MYNODE', [ax25.netrom.Destination('W1AW', 'NODE1', 'KU6S', 42)],
        b'\xffMYNODE'
        b'\xae\x62\x82\xae\x40\x40\x60NODE1 '
        b'\x96\xaa\x6c\xa6\x40\x40\x60\x2a'),
    ('MYNODE',
        [
            ax25.netrom.Destination('W1AW', 'NODE1', 'KU6S', 42),
            ax25.netrom.Destination('WR6ABD', 'NODE2', 'K6EAG', 21)
        ],
        b'\xffMYNODE'
        b'\xae\x62\x82\xae\x40\x40\x60NODE1 '
        b'\x96\xaa\x6c\xa6\x40\x40\x60\x2a'
        b'\xae\xa4\x6c\x82\x84\x88\x60NODE2 '
        b'\x96\x6c\x8a\x82\x8e\x40\x60\x15')
])
def test_pack(sender, destinations, expected):
    rb = ax25.netrom.RoutingBroadcast(sender, destinations)
    packed = rb.pack()
    assert packed == expected


@pytest.mark.parametrize(
    "in_packed, sender, destinations",
    [
        (
            b'\xffMYNODE',
            'MYNODE', None
        ),
        (
            b'\xffMYNODE'
            b'\xae\x62\x82\xae\x40\x40\x00NODE1 '
            b'\x96\xaa\x6c\xa6\x40\x40\x00\x2a',
            'MYNODE', [ax25.netrom.Destination('W1AW', 'NODE1', 'KU6S', 42)]
        ),
        (
            b'\xffMYNODE'
            b'\xae\x62\x82\xae\x40\x40\x00NODE1 '
            b'\x96\xaa\x6c\xa6\x40\x40\x00\x2a'
            b'\xae\xa4\x6c\x82\x84\x88\x00NODE2 '
            b'\x96\x6c\x8a\x82\x8e\x40\x00\x15',
            'MYNODE',
            [
                ax25.netrom.Destination('W1AW', 'NODE1', 'KU6S', 42),
                ax25.netrom.Destination('WR6ABD', 'NODE2', 'K6EAG', 21)
            ]
        )
    ])
def test_unpack(in_packed, sender, destinations):
    rb = ax25.netrom.RoutingBroadcast.unpack(in_packed)
    assert rb.sender == sender
    if destinations:
        assert len(rb.destinations) == len(destinations)
        dest_str = [str(d) for d in destinations]
        for d in rb.destinations:
            assert str(d) in dest_str
    else:
        assert rb.destinations is None


def test_unpack_bad():
    packed = b'\xeeMYNODE'
    with pytest.raises(TypeError):
        _ = ax25.netrom.RoutingBroadcast.unpack(packed)



# =============================================================================
# Copyright (c) 2020-2024 Martin F N Cooper
#
# Author: Martin F N Cooper
# License: MIT License
# =============================================================================

import pytest
import ax25


@pytest.mark.parametrize("in_call, in_base_only, result", [
    ('W1AW', False, True),
    ('W1AW-0', False, True),
    ('W1AW-1', False, True),
    ('W1AW-16', False, False),
    ('W1-AW-AW', False, False),
    ('W1!AW', False, False),
    ('W1AWAWAW', False, False),
    ('W1AW-32', False, False),
    ('W1AW-ABC', False, False),
    ('W1AW', True, True),
    ('W1AW-0', True, False),
    ('W1AW-1', True, False),
    ('W1AW-16', True, False),
    ('W1-AW-AW', True, False),
    ('W1!AW', True, False),
    ('W1AWAWAW', True, False),
    ('W1AW-32', True, False),
    ('W1AW-ABC', True, False)
])
def test_valid_call(in_call, in_base_only, result):
    assert ax25.Address.valid_call(in_call, in_base_only) == result


@pytest.mark.parametrize("in_call", [
    ('W1AW'),   # 1 x 2
    ('KU6S'),   # 2 x 1
    ('WW6HP'),  # 2 x 2
    ('K6EAG'),  # 1 x 3
    ('WR6ABD')  # 2 x 3
])
def test_callsign_patterns(in_call):
    addr = ax25.Address(in_call)
    assert addr._call == in_call
    assert addr._ssid == 0
    assert not addr._repeater
    assert not addr._command_response
    assert not addr._has_been_repeated


@pytest.mark.parametrize("in_call, call, ssid", [
    ('W1AW', 'W1AW', 0),
    ('W1AW-0', 'W1AW', 0),
    ('W1AW-1', 'W1AW', 1)
])
def test_callsign_only(in_call, call, ssid):
    addr = ax25.Address(in_call)
    assert addr._call == call
    assert addr._ssid == ssid
    assert not addr._repeater
    assert not addr._command_response
    assert not addr._has_been_repeated


@pytest.mark.parametrize("in_call, in_ssid, call, ssid", [
    ('W1AW', 1, 'W1AW', 1),
    ('W1AW-1', 1, 'W1AW', 1)
])
def test_callsign_and_ssid(in_call, in_ssid, call, ssid):
    addr = ax25.Address(in_call, in_ssid)
    assert addr._call == call
    assert addr._ssid == ssid
    assert not addr._repeater
    assert not addr._command_response
    assert not addr._has_been_repeated


@pytest.mark.parametrize("in_call, call, ssid, repeated", [
    ('W1AW', 'W1AW', 0, False),
    ('W1AW*', 'W1AW', 0, True),
    ('W1AW-2*', 'W1AW', 2, True)
])
def test_callsign_repeater(in_call, call, ssid, repeated):
    addr = ax25.Address(in_call, repeater=True)
    assert addr._call == call
    assert addr._ssid == ssid
    assert addr._repeater
    assert not addr._command_response
    assert addr._has_been_repeated == repeated


@pytest.mark.parametrize("in_call", [
    ('W1-AW-AW'),
    ('W1!AW'),
    ('W1AWAWAW'),
    ('W1AW-32'),
    ('W1AW-ABC')
])
def test_invalid_callsign(in_call):
    with pytest.raises(ValueError):
        _ = ax25.Address(in_call)


def test_ssid_conflict():
    with pytest.raises(ValueError):
        _ = ax25.Address("W1AW-2", 3)


def test_bad_repeater():
    with pytest.raises(TypeError):
        _ = ax25.Address("W1AW-1*")


@pytest.mark.parametrize("in_call, call", [
    ('W1AW', 'W1AW'),
    ('W1AW-0', 'W1AW'),
    ('W1AW-1', 'W1AW-1')
])
def test_str_callsign_only(in_call, call):
    assert str(ax25.Address(in_call)) == call


@pytest.mark.parametrize("in_call, in_ssid, call", [
    ('W1AW', 0, 'W1AW'),
    ('W1AW', 1, 'W1AW-1'),
    ('W1AW-1', 1, 'W1AW-1')
])
def test_str_callsign_and_ssid(in_call, in_ssid, call):
    assert str(ax25.Address(in_call, in_ssid)) == call


@pytest.mark.parametrize("in_call, call", [
    ('W1AW*', 'W1AW*'),
    ('W1AW-2*', 'W1AW-2*')
])
def test_str_callsign_only_repeated(in_call, call):
    assert str(ax25.Address(in_call, repeater=True)) == call


def test_str_callsign_and_ssid_repeated():
    assert str(ax25.Address("W1AW*", 2, True)) == "W1AW-2*"


@pytest.mark.parametrize(
    (
        "in_call, in_ssid, in_repeater, in_hbr, in_cr,"
        "rep_call, rep_ssid, rep_hbr, rep_cr"
    ),
    [
        ('W1AW', 0, False, None, None, 'W1AW', 0, None, False),
        ('W1AW', 1, False, None, None, 'W1AW', 1, None, False),
        ('W1AW-1', 0, False, None, None, 'W1AW', 1, None, False),
        ('W1AW', 2, False, None, True, 'W1AW', 2, None, True),
        ('W1AW', 0, True, None, None, 'W1AW', 0, False, None),
        ('W1AW', 1, True, None, None, 'W1AW', 1, False, None),
        ('W1AW*', 0, True, None, None, 'W1AW', 0, True, None),
        ('W1AW', 0, True, True, None, 'W1AW', 0, True, None)
    ])
def test_repr(
    in_call, in_ssid, in_repeater, in_hbr, in_cr,
        rep_call, rep_ssid, rep_hbr, rep_cr):
    addr = ax25.Address(in_call, in_ssid, in_repeater)
    if in_hbr is not None:
        addr.has_been_repeated = in_hbr
    if in_cr is not None:
        addr.command_response = in_cr
    rep = repr(addr)
    assert rep.startswith('Address(')
    assert rep.endswith(')')
    assert "call: {}".format(rep_call) in rep
    assert "ssid: {}".format(rep_ssid) in rep
    if in_repeater:
        assert "has_been_repeated: {}".format(rep_hbr) in rep
    else:
        assert "command_response: {}".format(rep_cr) in rep


def test_repeater_prop_unset():
    addr = ax25.Address("W1AW")
    assert not addr.repeater


@pytest.mark.parametrize("in_call, in_repeater, repeater", [
    ('W1AW', False, False),
    ('W1AW', True, True)
])
def test_repeater_prop_provided(in_call, in_repeater, repeater):
    addr = ax25.Address(in_call, repeater=in_repeater)
    assert addr.repeater == repeater


@pytest.mark.parametrize("in_call, has_been_repeated", [
    ('W1AW', False),
    ('W1AW*', True)
])
def test_repeated_prop(in_call, has_been_repeated):
    addr = ax25.Address(in_call, repeater=True)
    assert addr.has_been_repeated == has_been_repeated


def test_repeated_error():
    addr = ax25.Address("W1AW*", repeater=True)
    addr._repeater = False  # Field cannot be set via public access
    with pytest.raises(TypeError):
        _ = addr.has_been_repeated


@pytest.mark.parametrize("in_call, before, repeated, after", [
    ('W1AW', False, True, True),
    ('W1AW*', True, False, False)
])
def test_set_repeated(in_call, before, repeated, after):
    addr = ax25.Address(in_call, repeater=True)
    assert addr.has_been_repeated == before
    addr.has_been_repeated = repeated
    assert addr.has_been_repeated == after


def test_set_repeated_error():
    addr = ax25.Address("W1AW")
    with pytest.raises(TypeError):
        addr.has_been_repeated = True


def test_command_response_default():
    addr = ax25.Address("W1AW")
    assert not addr.command_response


def test_command_response_repeater():
    addr = ax25.Address("W1AW", repeater=True)
    with pytest.raises(TypeError):
        _ = addr.command_response


def test_set_command_response():
    addr = ax25.Address("W1AW")
    assert not addr.command_response
    addr.command_response = True
    assert addr.command_response


def test_set_command_response_repeater():
    addr = ax25.Address("W1AW", repeater=True)
    with pytest.raises(TypeError):
        addr.command_response = True


@pytest.mark.parametrize("in_call, packed_call", [
    ('WR6ABD', b'\xae\xa4\x6c\x82\x84\x88'),
    ('wr6abd', b'\xae\xa4\x6c\x82\x84\x88'),
    ('K6EAG', b'\x96\x6c\x8a\x82\x8e\x40'),
    ('KU6S', b'\x96\xaa\x6c\xa6\x40\x40')
])
def test_pack_callsign_only(in_call, packed_call):
    addr = ax25.Address(in_call)
    packed = addr.pack()
    assert len(packed) == 7
    assert packed[0:6] == packed_call
    assert packed[6] == ax25.RESERVED


@pytest.mark.parametrize("in_call, packed_call", [
    ('WR6ABD', b'\xae\xa4\x6c\x82\x84\x88'),
    ('wr6abd', b'\xae\xa4\x6c\x82\x84\x88'),
    ('K6EAG', b'\x96\x6c\x8a\x82\x8e\x40'),
    ('KU6S', b'\x96\xaa\x6c\xa6\x40\x40')
])
def test_pack_callsign_only_bytes(in_call, packed_call):
    addr = ax25.Address(in_call)
    packed = bytes(addr)
    assert len(packed) == 7
    assert packed[0:6] == packed_call
    assert packed[6] == ax25.RESERVED


@pytest.mark.parametrize("in_call, packed_call, last_byte", [
    ('W1AW-0', b'\xae\x62\x82\xae\x40\x40', 0x60),
    ('W1AW-1', b'\xae\x62\x82\xae\x40\x40', 0x62),
    ('W1AW-15', b'\xae\x62\x82\xae\x40\x40', 0x7E)
])
def test_pack_callsign_with_ssid(in_call, packed_call, last_byte):
    addr = ax25.Address(in_call)
    packed = addr.pack()
    assert len(packed) == 7
    assert packed[0:6] == packed_call
    assert packed[6] == last_byte


@pytest.mark.parametrize("in_call, in_ssid, packed_call, last_byte", [
    ('W1AW', 0, b'\xae\x62\x82\xae\x40\x40', 0x60),
    ('W1AW', 1, b'\xae\x62\x82\xae\x40\x40', 0x62),
    ('W1AW', 15, b'\xae\x62\x82\xae\x40\x40', 0x7E)
])
def test_pack_callsign_and_ssid(in_call, in_ssid, packed_call, last_byte):
    addr = ax25.Address(in_call, in_ssid)
    packed = addr.pack()
    assert len(packed) == 7
    assert packed[0:6] == packed_call
    assert packed[6] == last_byte


@pytest.mark.parametrize("in_call, packed_call, last_byte", [
    ('W1AW', b'\xae\x62\x82\xae\x40\x40', 0x60),
    ('W1AW*', b'\xae\x62\x82\xae\x40\x40', 0xE0),
    ('W1AW-3', b'\xae\x62\x82\xae\x40\x40', 0x66),
    ('W1AW-3*', b'\xae\x62\x82\xae\x40\x40', 0xE6)
])
def test_pack_has_been_repeated(in_call, packed_call, last_byte):
    addr = ax25.Address(in_call, repeater=True)
    packed = addr.pack()
    assert len(packed) == 7
    assert packed[0:6] == packed_call
    assert packed[6] == last_byte


@pytest.mark.parametrize("in_call, in_cmdresp, packed_call, last_byte", [
    ('W1AW', False, b'\xae\x62\x82\xae\x40\x40', 0x60),
    ('W1AW', True, b'\xae\x62\x82\xae\x40\x40', 0xE0),
    ('W1AW-3', False, b'\xae\x62\x82\xae\x40\x40', 0x66),
    ('W1AW-3', True, b'\xae\x62\x82\xae\x40\x40', 0xE6)
])
def test_pack_command_response(in_call, in_cmdresp, packed_call, last_byte):
    addr = ax25.Address(in_call)
    addr.command_response = in_cmdresp
    packed = addr.pack()
    assert len(packed) == 7
    assert packed[0:6] == packed_call
    assert packed[6] == last_byte


@pytest.mark.parametrize("in_packed, call, ssid, cmdresp", [
    (b'\xae\x62\x82\xae\x40\x40\x60', 'W1AW', 0, False),
    (b'\xae\x62\x82\xae\x40\x40\x66', 'W1AW', 3, False),
    (b'\xae\x62\x82\xae\x40\x40\xE0', 'W1AW', 0, True),
    (b'\xae\x62\x82\xae\x40\x40\xE6', 'W1AW', 3, True),
    (b'\x96\x6c\x8a\x82\x8e\x40\x64', 'K6EAG', 2, False),
    (b'\xae\xa4\x6c\x82\x84\x88\x64', 'WR6ABD', 2, False)
])
def test_unpack_non_repeater(in_packed, call, ssid, cmdresp):
    addr = ax25.Address.unpack(in_packed)
    assert addr.call == call
    assert addr.ssid == ssid
    assert not addr.repeater
    assert addr.command_response == cmdresp


@pytest.mark.parametrize("in_packed, call, ssid, cmdresp", [
    (b'\xae\x62\x82\xae\x40\x40\x00', 'W1AW', 0, False),
    (b'\xae\x62\x82\xae\x40\x40\x06', 'W1AW', 3, False),
    (b'\xae\x62\x82\xae\x40\x40\x80', 'W1AW', 0, True),
    (b'\xae\x62\x82\xae\x40\x40\x86', 'W1AW', 3, True),
    (b'\x96\x6c\x8a\x82\x8e\x40\x04', 'K6EAG', 2, False),
    (b'\xae\xa4\x6c\x82\x84\x88\x04', 'WR6ABD', 2, False)
])
def test_unpack_non_repeater_ignore_reserved(in_packed, call, ssid, cmdresp):
    addr = ax25.Address.unpack(in_packed)
    assert addr.call == call
    assert addr.ssid == ssid
    assert not addr.repeater
    assert addr.command_response == cmdresp


def test_unpack_non_repeater_repeated():
    addr = ax25.Address.unpack(b'\x96\x88\x6c\xb2\x82\x9a\x00')
    with pytest.raises(TypeError):
        assert addr.has_been_repeated


@pytest.mark.parametrize("in_packed, call, ssid, repeated", [
    (b'\xae\x62\x82\xae\x40\x40\x60', 'W1AW', 0, False),
    (b'\xae\x62\x82\xae\x40\x40\x66', 'W1AW', 3, False),
    (b'\xae\x62\x82\xae\x40\x40\xE0', 'W1AW', 0, True),
    (b'\xae\x62\x82\xae\x40\x40\xE6', 'W1AW', 3, True),
    (b'\x96\x6c\x8a\x82\x8e\x40\x64', 'K6EAG', 2, False),
    (b'\xae\xa4\x6c\x82\x84\x88\x64', 'WR6ABD', 2, False)
])
def test_unpack_repeater(in_packed, call, ssid, repeated):
    addr = ax25.Address.unpack(in_packed, repeater=True)
    assert addr.call == call
    assert addr.ssid == ssid
    assert addr.repeater
    assert addr.has_been_repeated == repeated


@pytest.mark.parametrize("in_packed, call, ssid, repeated", [
    (b'\xae\x62\x82\xae\x40\x40\x00', 'W1AW', 0, False),
    (b'\xae\x62\x82\xae\x40\x40\x06', 'W1AW', 3, False),
    (b'\xae\x62\x82\xae\x40\x40\x80', 'W1AW', 0, True),
    (b'\xae\x62\x82\xae\x40\x40\x86', 'W1AW', 3, True),
    (b'\x96\x6c\x8a\x82\x8e\x40\x04', 'K6EAG', 2, False),
    (b'\xae\xa4\x6c\x82\x84\x88\x04', 'WR6ABD', 2, False)
])
def test_unpack_repeater_ignore_reserved(in_packed, call, ssid, repeated):
    addr = ax25.Address.unpack(in_packed, repeater=True)
    assert addr.call == call
    assert addr.ssid == ssid
    assert addr.repeater
    assert addr.has_been_repeated == repeated


def test_unpack_repeater_cmdresp():
    addr = ax25.Address.unpack(b'\xae\x62\x82\xae\x40\x40\x00', True)
    with pytest.raises(TypeError):
        assert addr.command_response


def test_unpack_invalid_callsign():
    with pytest.raises(ValueError):
        _ = ax25.Address.unpack(b'\xae\x42\x82\xae\x40\x40\x00')



# =============================================================================
# Copyright (c) 2020-2024 Martin F N Cooper
#
# Author: Martin F N Cooper
# License: MIT License
# =============================================================================

"""
NET/ROM Routing Table Update Structure

Structure definitions together with methods to pack and unpack native NET/ROM
routing table updates. Provides the capabilities for, for example, decoding
NET/ROM routing table update packets in programs such as 'listen'.

Protocol reference:
  https://packet-radio.net/wp-content/uploads/2017/04/netrom1.pdf
"""

import ax25

# Destination offsets
OFF_CALLSIGN = 0
OFF_MNEMONIC = ax25.AXLEN
OFF_NEIGHBOR = ax25.AXLEN + ax25.ALEN
OFF_QUALITY  = ax25.AXLEN + ax25.ALEN + ax25.AXLEN

MNEM_LEN = 6
DEST_SIZE = ax25.AXLEN + MNEM_LEN + ax25.AXLEN + 1


class Destination:
    """
    A single destination element of a routing update.

    This comprises identifiers for the node, along with information on its
    best-quality neighbor.

    :param callsign: Callsign of the destination.
    :type callsign: Address or str
    :param str mnemonic: Mnemonic for the destination.
    :param neighbor: Best-quality neighbor for the destination.
    :type neighbor: Address or str
    :param int quality: Quality of the best-quality neighbor.
    :raises TypeError: if an address is of an invalid type.
    :raises ValueError: if either the mnemonic value or the quality value is
        invalid or out of range.
    """
    def __init__(self, callsign, mnemonic, neighbor, quality):
        # This is copied from Frame.check_subfield; need a common function
        def check_address(addr):
            if isinstance(addr, ax25.Address):
                return addr
            elif isinstance(addr, str):
                return ax25.Address(addr)
            else:
                raise TypeError('Invalid address field')
        self._callsign = check_address(callsign)
        self._neighbor = check_address(neighbor)
        if not (isinstance(mnemonic, str) and 0 < len(mnemonic) <= MNEM_LEN):
            raise ValueError('Invalid mnemonic')
        self._mnemonic = mnemonic
        if not (isinstance(quality, int) and 0 <= quality <= 255):
            raise ValueError('Invalid quality value')
        self._quality = quality

    def __repr__(self):
        return ('Destination(callsign: {}, mnemonic: {}, neighbor: {},'
                ' quality: {})'.format(
                    self._callsign, self._mnemonic, self._neighbor,
                    self._quality))

    def __str__(self):
        return '{} ({}) -> {} ({})'.format(
            self._callsign, self._mnemonic, self._neighbor, self._quality)

    @property
    def callsign(self):
        """
        Retrieve the callsign of the destination. Readonly.

        :returns: Destination callsign.
        :rtype: Address
        """
        return self._callsign

    @property
    def mnemonic(self):
        """
        Retrieve the mnemonic for the destination. Readonly.

        :returns: Destination's mnemonic.
        :rtype: str
        """
        return self._mnemonic

    @property
    def best_neighbor(self):
        """
        Retrieve the best-quality neighbor for the destination. Readonly.

        :returns: The best-quality neighbor.
        :rtype: Address
        """
        return self._neighbor

    @property
    def best_quality(self):
        """
        Retrieve the quality of the best-quality neighbor. Readonly.

        :returns: The best-quality neighbor's quality.
        :rtype: int
        """
        return self._quality

    def pack(self):
        """
        Pack this :class:`Destination` instance into an encoded byte sequence.

        :returns: Encoded byte sequence for this :class:`Destination`.
        :rtype: bytes
        """
        b = bytearray(DEST_SIZE)
        b[OFF_CALLSIGN:OFF_CALLSIGN + ax25.AXLEN] = self._callsign.pack()
        b[OFF_MNEMONIC:OFF_MNEMONIC + ax25.ALEN] = \
            self._mnemonic.ljust(ax25.ALEN).encode('utf-8')
        b[OFF_NEIGHBOR:OFF_NEIGHBOR + ax25.AXLEN] = self._neighbor.pack()
        b[OFF_QUALITY] = self._quality
        return bytes(b)

    @classmethod
    def unpack(cls, buffer):
        """
        Unpack the encoded byte sequence into a new :class:`Destination`
        instance.

        :param buffer: Encoded byte sequence.
        :type buffer: bytes or bytearray
        :returns: A new Destination instance.
        :rtype: Destination
        :raises ValueError: if the encoded information contains an invalid
            address.
        """
        callsign = ax25.Address.unpack(
            buffer[OFF_CALLSIGN:OFF_CALLSIGN + ax25.AXLEN])
        mnemonic = bytes(
            buffer[OFF_MNEMONIC:OFF_MNEMONIC + ax25.ALEN]).decode(
                'utf-8').rstrip()
        neighbor = ax25.Address.unpack(
            buffer[OFF_NEIGHBOR:OFF_NEIGHBOR + ax25.AXLEN])
        quality = buffer[OFF_QUALITY]
        return cls(callsign, mnemonic, neighbor, quality)


class RoutingBroadcast:
    """
    A complete routing update broadcast.

    This comprises sender identification along with a list of destinations.

    :param str sender: Mnemonic of the sender.
    :param destinations: List or tuple of destinations.
    :type destinations: list[Destination] or tuple(Destination)
    :raises TypeError: if a destination is of an invalid type.
    :raises ValueError: if the sender is invalid.
    """
    def __init__(self, sender, destinations):
        if not (isinstance(sender, str) and 0 < len(sender) <= MNEM_LEN):
            raise ValueError('Invalid sender')
        self._sender = sender
        if destinations:
            if not isinstance(destinations, (list, tuple)):
                raise TypeError('Invalid destinations')
            for d in destinations:
                if not isinstance(d, Destination):
                    raise TypeError('Invalid destinations')
            self._destinations = tuple(destinations)
        else:
            self._destinations = None

    def __repr__(self):
        if self._destinations:
            destinations = ','.join(map(repr, self._destinations))
            rep = ('RoutingBroadcast(sender: {},'
                   ' destinations: [{}])'.format(self._sender, destinations))
        else:
            rep = 'RoutingBroadcast(sender: {})'.format(self._sender)
        return rep

    @property
    def sender(self):
        """
        Retrieve the sender of the routing broadcast. Readonly.

        :returns: Sender of the routing broadcast.
        :rtype: str
        """
        return self._sender

    @property
    def destinations(self):
        """
        Retrieve the destinations for routing broadcast. Readonly.

        :returns: Sequence of destinations.
        :rtype: tuple(Destination)
        """
        return self._destinations

    def pack(self):
        """
        Pack this :class:`RoutingBroadcast` instance into an encoded byte
        sequence.

        :returns: Encoded byte sequence for this :class:`RoutingBroadcast`.
        :rtype: bytes
        """
        pack_len = 1 + ax25.ALEN
        if self._destinations:
            pack_len += len(self._destinations) * DEST_SIZE
        b = bytearray(pack_len)
        b[0] = 0xFF
        b[1:1 + ax25.ALEN] = self._sender.ljust(ax25.ALEN).encode('utf-8')
        offset = 1 + ax25.ALEN
        if self._destinations:
            for dest in self._destinations:
                b[offset:offset + DEST_SIZE] = dest.pack()
                offset += DEST_SIZE
        return bytes(b)

    @classmethod
    def unpack(cls, buffer):
        """
        Unpack the encoded byte sequence into a new :class:`RoutingBroadcast`
        instance.

        :param buffer: Encoded byte sequence.
        :type buffer: bytes or bytearray
        :returns: A new RoutingBroadcast instance.
        :rtype: RoutingBroadcast
        :raises TypeError: if the byte sequence does not represent a routing
            broadcast (i.e. if the first byte is not 0xFF).
        :raises ValueError: if the encoded information contains an invalid
            address.
        """
        if buffer[0] != 0xFF:
            raise TypeError("Not a routing broadcast")
        sender = bytes(buffer[1:1 + ax25.ALEN]).decode('utf-8')
        destinations = []
        remaining = len(buffer) - ax25.ALEN - 1
        offset = ax25.ALEN + 1
        while remaining:
            destination = Destination.unpack(
                buffer[offset:offset + DEST_SIZE])
            destinations.append(destination)
            offset += DEST_SIZE
            remaining -= DEST_SIZE
        return cls(sender, destinations)



# =============================================================================
# Copyright (c) 2018-2024 Martin F N Cooper
#
# Author: Martin F N Cooper
# License: MIT License
# =============================================================================

"""
AX.25 Frame Structure

Structure definitions together with methods to pack and unpack native AX.25
frames and the constituent parts of those frames. This can be used together
with a communication protocol such as AGWPE, KISS, or a native AX.25 stack
for applications across the gamut, from simple 'listen' monitoring to full
connected mode systems.

Protocol reference:
  http://www.tapr.org/pdf/AX25.2.2.pdf
"""

__author__ = 'Martin F N Cooper'
__version__ = '1.0.1'

from enum import Enum

# Addressing
ALEN = 6             # Basic address length (callsign only)
AXLEN = 7            # Extended address length (with SSID)
OFF_DST = 0          # Offset to destination address
OFF_SRC = AXLEN      # Offset to source address
OFF_VIA = AXLEN * 2  # Offset to 'via' fields

# SSID byte
HDLC_AEB    = 0x01  # HDLC address extension bit
SSID        = 0x1E  # SSID mask
CMDRESP     = 0x80  # Command / response
REPEATED    = 0x80  # Has been repeated
RESERVED    = 0x60  # Reserved bits
SSSID_SPARE = 0x40  # EAX25
ESSID_SPARE = 0x20  # DAMA

# Frame control byte
PF  = 0x10  # P/F
NR  = 0xE0  # N(R) mask
NRS = 5     # N(R) shift
NS  = 0x0E  # N(S) mask
NSS = 1     # N(S) shift


class FrameType(Enum):
    """
    The type of an AX.25 frame, as identified within the control field.

    The values match those defined in the AX.25 spec, with an additional value
    to identify an as yet unknown frame. General types are also defined, to
    allow for categorization of I, S and U frames, along with methods to
    determine whether a value matches a specific category.
    """
    # S frame types
    RR    = 0x01
    """ Receive Ready """
    RNR   = 0x05
    """ Receive Not Ready """
    REJ   = 0x09
    """ Reject """
    SREJ  = 0x0D
    """ Selective Reject """
    # U frame types
    UI    = 0x03
    """ Unnumbered Information """
    DM    = 0x0F
    """ Disconnect Mode """
    SABM  = 0x2F
    """ Set Async Balanced Mode """
    DISC  = 0x43
    """ Disconnect """
    UA    = 0x63
    """ Unnumbered Acknowledge """
    SABME = 0x6F
    """ Set Async Balanced Mode """
    FRMR  = 0x87
    """ Frame Reject """
    XID   = 0xAF
    """ Exchange Identification """
    TEST  = 0xE3
    """ Test """
    # Unknown frame type
    UNK   = 0xFF
    """ Unknown frame type """
    # General frame types
    I     = 0x00  # noqa: E741
    """ I Frame """
    S     = 0x01
    """ S Frame """
    U     = 0x03
    """ U Frame """

    def is_I(self):  # noqa N802
        """ Is this an I frame type? """
        return self.value & 0x01 == 0x00

    def is_S(self):  # noqa N802
        """ Is this an S frame type? """
        return self.value & 0x03 == 0x01

    def is_U(self):  # noqa N802
        """ Is this an U frame type? """
        return self.value & 0x03 == 0x03


class Address:
    """
    A single element (subfield) of the address field for an AX.25 frame.

    In addition to the callsign and SSID, this includes fields corresponding
    to has-been-repeated and command/response state.

    :param str call: Callsign, with or without an SSID. If an SSID is included
        in this string, the `ssid` argument must not be used. If the callsign
        ends with an asterisk, the address is marked as having been repeated.
    :param int ssid: SSID, in the range 0..15.
    :param bool repeater: Indicates whether or not this instance represents a
        repeater address. This is used to determine whether this instance has
        a command/response bit or a has-been-repeated bit.
    :raises TypeError: if the `repeater` argument is `False` but the callsign
        ends with an asterisk.
    :raises ValueError: if the callsign or SSID is invalid, or if the callsign
        includes an SSID value that conflicts with the `ssid` argument.
    """
    def __init__(self, call, ssid=0, repeater=False):
        self._repeater = repeater
        self._command_response = False
        self._has_been_repeated = False
        if call.endswith('*'):
            if repeater:
                self._has_been_repeated = True
                call = call[:-1]
            else:
                raise TypeError(
                    'Cannot set has_been_repeated on non-repeater address')
        parts = call.split('-')
        if len(parts) > 2 or not self.valid_call(parts[0], True):
            raise ValueError('Invalid callsign')
        if len(parts) == 2:
            ssid2 = int(parts[1])
            if ssid != 0 and ssid2 != ssid:
                raise ValueError('SSID conflict')
            ssid = ssid2
        if ssid > 15:
            raise ValueError('Invalid SSID')
        self._call = parts[0].upper()
        self._ssid = ssid

    def __repr__(self):
        return ('Address(call: {}, ssid: {}, repeater: {}, {}: {})').format(
            self._call, self._ssid, self._repeater,
            ('has_been_repeated' if self._repeater
                else 'command_response'),
            (self._has_been_repeated if self._repeater
                else self._command_response))

    def __str__(self):
        call = self._call
        if self._ssid:
            call += '-' + str(self._ssid)
        if self._repeater and self._has_been_repeated:
            call += '*'
        return call

    def __bytes__(self):
        return self.pack()

    @staticmethod
    def valid_call(call, base_only=False):
        """
        Check a callsign to determine its validity.

        The following checks are performed:

         - Base callsign is of a valid length, and is alphanumeric.
         - Zero or one callsign/SSID separators are present.
         - SSID, if present, is numeric and within the valid range.

        :param str call: The callsign to be validated.
        :param bool base_only: If `True`, the callsign will be validated as
            one without an SSID, such that a callsign *with* an SSID will be
            considered invalid; if `False`, an SSID may be present.
        :returns: `True` if the callsign is valid; `False` otherwise.
        :rtype: bool
        """
        if not base_only:
            parts = call.split('-')
            if len(parts) > 2:
                return False
            if len(parts) == 2:
                if not parts[1].isdigit():
                    return False
                ssid = int(parts[1])
                if ssid > 15:
                    return False
                call = parts[0]
        return 2 <= len(call) <= ALEN and call.isalnum()

    @property
    def call(self):
        """
        Retrieve the callsign (only) for this address. Readonly.

        :returns: The callsign, without SSID.
        :rtype: str
        """
        return self._call

    @property
    def ssid(self):
        """
        Retrieve the SSID for this address. Readonly.

        :returns: The SSID.
        :rtype: int
        """
        return self._ssid

    @property
    def repeater(self):
        """
        Determine whether or not this is a repeater address. Readonly.

        :returns: `True` if this is a repeater address; `False` otherwise.
        :rtype: bool
        """
        return self._repeater

    @property
    def has_been_repeated(self):
        """
        Determine whether or not this repeater address has been repeated.

        The getter returns `True` if this address has been repeated; `False`
        otherwise.

        The setter takes a `bool` value, and raises :class:`TypeError` if an
        attempt is made to set this property for a non-repeater address.
        """
        if not self._repeater:
            raise TypeError(
                'has_been_repeated not valid for non-repeater address')
        return self._has_been_repeated

    @has_been_repeated.setter
    def has_been_repeated(self, value):
        if not self._repeater:
            raise TypeError(
                'Cannot set has_been_repeated on non-repeater address')
        self._has_been_repeated = value

    @property
    def command_response(self):
        """
        Determine whether or not the command/response bit is set for this
        non-repeater address.

        The getter returns `True` if the command/response bit is set; `False`
        otherwise.

        The setter takes a `bool` value, and raises :class:`TypeError` if an
        attempt is made to set this property for a repeater address.
        """
        if self._repeater:
            raise TypeError(
                'command_response not valid for repeater address')
        return self._command_response

    @command_response.setter
    def command_response(self, value):
        if self._repeater:
            raise TypeError('Cannot set command_response on repeater address')
        self._command_response = value

    def pack(self):
        """
        Pack this :class:`Address` instance into an encoded byte sequence.

        :returns: Encoded byte sequence for this :class:`Address`.
        :rtype: bytes
        """
        b = bytearray(AXLEN)
        for i, c in enumerate(self._call):
            b[i] = ord(c) << 1
        for p in range(i + 1, ALEN):
            b[p] = ord(' ') << 1
        b[ALEN] = self._ssid << 1
        if self._repeater:
            if self._has_been_repeated:
                b[ALEN] |= REPEATED
        else:
            if self._command_response:
                b[ALEN] |= CMDRESP
        b[ALEN] |= RESERVED
        return bytes(b)

    @classmethod
    def unpack(cls, buffer, repeater=False):
        """
        Unpack the encoded byte sequence into a new :class:`Address` instance.

        The `repeater` argument specifies whether the byte sequence represents
        a repeater or non-repeater address.

        :param buffer: Encoded byte sequence.
        :type buffer: bytes or bytearray
        :param bool repeater: Whether or not this address represents a
            repeater address.
        :returns: A new Address instance.
        :rtype: Address
        :raises ValueError: if the encoded information results in an invalid
            callsign.
        """
        call = ""
        for i in range(0, ALEN):
            b = buffer[i]
            c = (b >> 1) & 0x7F
            if c != ord(' '):
                call += chr(c)
        if not cls.valid_call(call, True):
            raise ValueError('Invalid callsign')
        ssid = (buffer[ALEN] & SSID) >> 1
        addr = cls(call, ssid, repeater)
        if repeater:
            addr.has_been_repeated = (buffer[ALEN] & REPEATED) != 0
        else:
            addr.command_response = (buffer[ALEN] & CMDRESP) != 0
        return addr


class Control:
    """
    The control field for an AX.25 frame.

    This includes the frame type, the poll/final indicator, and the send and
    receive sequence numbers. At this time, only modulo 8 (single octet)
    control fields are supported.

    :param FrameType frame_type: Type of the associated :class:`Frame`.
    :param bool poll_final: State of the poll/final indicator.
    :param int recv_seqno: Receive sequence number.
    :param int send_seqno: Send sequence number.
    """
    def __init__(
            self, frame_type, poll_final=False, recv_seqno=0, send_seqno=0):
        self._frame_type = frame_type
        self._poll_final = poll_final
        self._recv_seqno = recv_seqno
        self._send_seqno = send_seqno

    def __repr__(self):
        fmt = 'Control(type: {}, poll_final: {}'
        args = [self._frame_type.name, self._poll_final]
        if self._frame_type.is_I() or self._frame_type.is_S():
            fmt += ', recv_seqno: {}'
            args += str(self._recv_seqno)
        if self._frame_type.is_I():
            fmt += ', send_seqno: {}'
            args += str(self._send_seqno)
        fmt += ')'
        return fmt.format(*args)

    def __str__(self):
        control = self.pack()
        return '{:02X}'.format(control)

    def __int__(self):
        return self.pack()

    @property
    def frame_type(self):
        """
        Retrieve the type of the associated frame. Readonly.

        :returns: Type of the frame.
        :rtype: FrameType
        """
        return self._frame_type

    @property
    def poll_final(self):
        """
        Determine whether or not the poll/final bit is set.

        The getter returns `True` if the poll/final bit is set; `False`
        otherwise.

        The setter takes a `bool` value to set or unset the poll/final bit.
        """
        return self._poll_final

    @poll_final.setter
    def poll_final(self, v):
        self._poll_final = v

    @property
    def recv_seqno(self):
        """
        Retrieve or set the receive sequence number.

        The receive sequence number is an `int`, and is valid only on I and S
        frames; a :class:`TypeError` is raised if an attempt is made to
        retrieve or set this value on a U frame.
        """
        if self._frame_type.is_U():
            raise TypeError('recv_seqno not valid on U frame')
        return self._recv_seqno

    @recv_seqno.setter
    def recv_seqno(self, v):
        if self._frame_type.is_U():
            raise TypeError('Cannot set recv_seqno on U frame')
        self._recv_seqno = v

    @property
    def send_seqno(self):
        """
        Retrieve or set the send sequence number.

        The send sequence number is an `int`, and is valid only on I frames;
        a :class:`TypeError` is raised if an attempt is made to retrieve or
        set this value on an S or U frame.
        """
        if self._frame_type.is_S():
            raise TypeError('send_seqno not valid on S frame')
        if self._frame_type.is_U():
            raise TypeError('send_seqno not valid on U frame')
        return self._send_seqno

    @send_seqno.setter
    def send_seqno(self, v):
        if self._frame_type.is_S():
            raise TypeError('Cannot set send_seqno on S frame')
        if self._frame_type.is_U():
            raise TypeError('Cannot set send_seqno on U frame')
        self._send_seqno = v

    def pack(self):
        """
        Pack this :class:`Control` instance into an encoded value.

        :returns: Encoded value for this :class:`Control`.
        :rtype: int
        :raises ValueError: if the frame type is not set
            (i.e. it is :py:const:`FrameType.UNK`).
        """
        if self._frame_type is FrameType.UNK:
            raise ValueError('Unknown frame type')
        control = self._frame_type.value
        if self._poll_final:
            control |= PF
        if self._frame_type.is_I() or self._frame_type.is_S():
            control |= (self._recv_seqno << NRS)
        if self._frame_type.is_I():
            control |= (self._send_seqno << NSS)
        return control

    @classmethod
    def unpack(cls, control):
        """
        Unpack the encoded value into a new :class:`Control` instance.

        Values for the poll/final bit and receive and send sequence numbers are
        determined based upon the frame type.

        :param int control: Encoded value.
        :returns: A new Control instance.
        :rtype: Control
        :raises ValueError: if the frame type is invalid.
        """
        # Determine frame type first
        ft = FrameType.UNK
        if (control & 0x01) == 0:  # I frame
            ft = FrameType.I
        elif (control & 0x02) == 0:  # S frame
            # May raise ValueError if type is invalid
            ft = FrameType(control & 0x0F)
        else:  # U frame
            # May raise ValueError if type is invalid
            ft = FrameType(control & ~PF)
        # Remaining fields, based on frame type
        pf = (control & PF) != 0
        if ft.is_I() or ft.is_S():
            nr = (control & NR) >> NRS
        else:
            nr = 0
        if ft.is_I():
            ns = (control & NS) >> NSS
        else:
            ns = 0
        return cls(ft, pf, nr, ns)


class Frame:
    """
    A complete AX.25 frame, excluding the flag and FCS fields.

    This includes the address field and control field, and the protocol
    identifier and information field where present.

    :param dst: Destination address. Required.
    :type dst: Address or str
    :param src: Source address. Required.
    :type src: Address or str
    :param via: List of Via addresses. Optional.
    :type via: list[Address or str] or None
    :param Control control: Control field. Required.
    :param int pid: Protocol Identifier. Valid only for I and UI frames.
    :param data: Data for information field. Valid only for I, UI, FRMR, XID
        and TEST frames.
    :type data: bytes or bytearray or None
    :raises TypeError: if any provided field has an invalid type.
    :raises ValueError: if data is provided for an invalid frame type, or if
        the data field is too long.
    """

    def __init__(self, dst, src, via=None, control=None, pid=0, data=None):
        # Addressing first
        def check_subfield(addr):
            if isinstance(addr, Address):
                return addr
            elif isinstance(addr, str):
                return Address(addr)
            else:
                raise TypeError('Invalid address field')
        self._dst = check_subfield(dst)
        self._src = check_subfield(src)
        if via:
            if not isinstance(via, (list, tuple)):
                raise TypeError('Invalid via')
            self._via = tuple(check_subfield(v) for v in via)
        else:
            self._via = None
        # Control field
        if isinstance(control, Control):
            self._control = control
        elif isinstance(control, int):
            self._control = Control.unpack(control)
        else:
            raise TypeError('Invalid control field')
        # Protocol identifier, if present
        ft = self._control.frame_type
        if self._has_pid(ft):
            self._pid = pid
        else:
            self._pid = 0
        # Data, if present
        if self._info_allowed(ft):
            if data is not None:
                if not isinstance(data, (bytearray, bytes)):
                    raise TypeError('Data field must be bytes or bytearray')
                elif len(data) > 256:
                    raise ValueError('Data field too long')
        elif data is not None:
            raise ValueError('Data field not valid for frame type')
        self._data = data

    def __bytes__(self):
        return self.pack()

    @staticmethod
    def _has_pid(ft):
        return ft is FrameType.I or ft is FrameType.UI

    @staticmethod
    def _info_allowed(ft):
        return ft in (
            FrameType.I,
            FrameType.UI,
            FrameType.FRMR,
            FrameType.XID,
            FrameType.TEST)

    @property
    def dst(self):
        """
        Retrieve the destination address for this frame. Readonly.

        :returns: Destination address.
        :rtype: Address
        """
        return self._dst

    @property
    def src(self):
        """
        Retrieve the source address for this frame. Readonly.

        :returns: Source address.
        :rtype: Address
        """
        return self._src

    @property
    def via(self):
        """
        Retrieve the list of Via addresses for this frame. Readonly.

        :returns: List of Via addresses.
        :rtype: list[Address] or None
        """
        return self._via

    @property
    def control(self):
        """
        Retrieve the control field for this frame. Readonly.

        :returns: Control field.
        :rtype: int
        """
        return self._control

    @property
    def pid(self):
        """
        Retrieve the Protocol Identifier for this frame. Readonly.

        :returns: Protocol identifier, or 0 if PID is not permitted for this
            frame type.
        :rtype: int
        """
        # Value will already be zero if pid not allowed for frame type
        return self._pid

    @property
    def data(self):
        """
        Retrieve the information field for this frame. Readonly.

        :returns: Information field data, or None if the information field is
            not permitted for this frame type.
        :rtype: bytes or None
        """
        # Value will already be None if data not allowed for frame type
        return self._data

    def pack(self):
        """
        Pack this :class:`Frame` instance into an encoded byte sequence.

        :returns: Encoded byte sequence for this :class:`Frame`.
        :rtype: bytes
        """
        b = bytearray()
        # Basic address fields
        b.extend(self._dst.pack())
        b.extend(self._src.pack())
        # Repeater addresses
        if self._via:
            for v in self._via:
                b.extend(v.pack())
        b[-1] |= HDLC_AEB
        # Control byte
        b.append(self._control.pack())
        # Protocol identifier, if required
        ft = self._control.frame_type
        if self._has_pid(ft):
            b.append(self._pid)
        if self._info_allowed(ft):
            if self._data:
                b.extend(self._data)
        return bytes(b)

    @classmethod
    def unpack(cls, buffer):
        """
        Unpack the encoded byte sequence into a new :class:`Frame` instance.

        :param buffer: Encoded byte sequence.
        :type buffer: bytes or bytearray
        :returns: A new Frame instance.
        :rtype: Frame
        :raises ValueError: if any encoded address results in an invalid
            callsign.
        """
        # Basic address fields
        dst = Address.unpack(buffer[OFF_DST:OFF_DST + AXLEN])
        src = Address.unpack(buffer[OFF_SRC:OFF_SRC + AXLEN])
        # Repeater addresses
        via = []
        offset = OFF_VIA
        end = (buffer[AXLEN + ALEN] & HDLC_AEB)
        if not end:
            while not end:
                via.append(
                    Address.unpack(buffer[offset:offset + AXLEN], True))
                end = buffer[offset + ALEN] & HDLC_AEB
                offset += AXLEN
        # Control byte
        control = Control.unpack(buffer[offset])
        offset = offset + 1
        # PID, if there is one
        ft = control.frame_type
        if cls._has_pid(ft):
            pid = buffer[offset]
            offset = offset + 1
        else:
            pid = 0
        # Data, if there is any
        if cls._info_allowed(ft):
            data = buffer[offset:]
        else:
            data = None

        return cls(dst, src, via, control, pid, data)
